from pydantic import BaseModel, Field, validator, ConfigDict, RootModel, field_validator, constr, model_validator, HttpUrl, conint
from typing import Optional, Literal, List, Dict, Union, Any, Type
import asyncio
import os
import uuid
import time
import torch
import re
import base64
from .video import VideoFrame


class TextTag(BaseModel):
    tag_type: Literal["text"] = "text"


class ArtifactTag(BaseModel):
    tag_type: Literal["artifact"] = "artifact"
    artifact_type: Literal["code", "self_contained_text"]
    identifier: str
    title: str
    language: Optional[str] = None


class Query(BaseModel):
    prompt: str
    temperature: Optional[float] = 0.1
    stream: Optional[bool] = False


class FunctionCall(BaseModel):
    arguments: str
    name: str

class ToolCall(BaseModel):
    id: str
    function: FunctionCall
    type: str = "function"
    index: Optional[int] = None


class MultiModalTextContent(BaseModel):
    type: Literal["text"]
    text: str = ""


class MultiModalImageContent(BaseModel):
    class ImageDetail(BaseModel):
        url: str = Field(description="URL, file path, or Data URI of the image")
        detail: Optional[Literal["low", "high"]] = "high"

        @validator('url')
        def validate_image_url(cls, v):
            if v.startswith('data:image/'):
                return cls.validate_data_uri(v)
            elif v.startswith('file://'):
                return cls.validate_local_file(v)
            elif v.startswith(('http://', 'https://')):
                return cls.validate_http_url(v)
            else:
                raise ValueError("Invalid image reference format")

        @classmethod
        def validate_data_uri(cls, v):
            data_uri_pattern = r'^data:image/(\w+);base64,(.+)$'
            match = re.match(data_uri_pattern, v)
            if not match:
                raise ValueError("Invalid Data URI format")

            image_format, base64_data = match.groups()
            try:
                decoded = base64.b64decode(base64_data)
                headers = {
                    'jpeg': b'\xff\xd8\xff',
                    'png': b'\x89PNG\r\n\x1a\n',
                    'gif': b'GIF87a',
                    'gif': b'GIF89a',
                    'webp': b'RIFF'
                }
                if not any(decoded.startswith(header) for header in headers.values()):
                    raise ValueError("Invalid image format")
            except base64.binascii.Error:
                raise ValueError("Invalid base64 string")
            return v

        @classmethod
        def validate_local_file(cls, v):
            file_path = v[7:]  # Remove 'file://' prefix
            if not os.path.isfile(file_path):
                raise ValueError("File does not exist")
            valid_extensions = ['.jpg', '.jpeg', '.png', '.gif', '.webp']
            if not any(file_path.lower().endswith(ext) for ext in valid_extensions):
                raise ValueError("Invalid image file extension")
            return v

        @classmethod
        def validate_http_url(cls, v):
            try:
                HttpUrl(v)
            except ValueError:
                raise ValueError("Invalid HTTP URL")
            return v

    type: Literal["image_url"]
    image_url: ImageDetail



class BaseMessage(BaseModel):
    role: Literal['system', 'user', 'assistant', 'tool']
    content: Optional[Union[str, List[Union[MultiModalTextContent, MultiModalImageContent]]]] = ""
    tool_calls: Optional[List[ToolCall]] = None
    tool_call_id: Optional[str] = None

    @validator('content', pre=True)
    def convert_null_to_empty(cls, value):
        if value is None:
            return ""
        return value


class ParameterProperties(BaseModel):
    type: str = "object"
    description: Optional[str] = None
    enum: Optional[List[str]] = None
    items: Optional[dict] = None   # for array


class ParameterSpec(BaseModel):
    type: str = "object"
    properties: Optional[Dict[str, Any]] = None
    required: List[str] = []


class FunctionSpec(BaseModel):
    name: str
    description: Optional[str] = None
    parameters: ParameterSpec = None
    strict: bool = True     # not making any difference at the moment


class ToolSpec(BaseModel):
    type: str = "function"
    function: FunctionSpec


class SingleFunctionDict(BaseModel):
    name: str


class ToolForce(BaseModel):
    type: str = "function"
    function: SingleFunctionDict

    class Config:
        extra = "forbid"  # This will prevent extra keys in the dictionary


# test_thinking = """
# <plan>
#   <task>Brief task summary</task>
#   <structure>
#     <c1>[text]: Acknowledge question and introduce answer</c1>
#     <c2>[artifact]: Main content (e.g., code)</c2>
#     <c3>[text]: Explain or elaborate on c2</c3>
#     <c4>[artifact]: Additional content if needed</c4>
#     <c5>[text]: Explain or elaborate on c4</c5>
#     <!-- Add more pairs if needed -->
#   </structure>
# </plan>
# """

class ChatMLQuery(BaseModel):
    # Configure the model to forbid extra fields
    model_config = ConfigDict(extra="forbid")

    class ResponseFormat(BaseModel):
        type: Literal["text", "json_object"]

    class StreamOption(BaseModel):
        include_usage: bool = False

    model: Optional[str] = ""
    messages: List[BaseMessage]
    temperature: Optional[float] = 0.01
    top_p: float = 0.85
    stream: Optional[bool] = False
    tools: Optional[List[ToolSpec]] = None
    tool_choice: Union[None, Literal["none", "auto", "required"], ToolForce] = None
    tool_call_id: Optional[str] = None
    max_tokens: Optional[int] = None

    # for video, currently for websocket
    video: Optional[List[Any]] = None    # TODO to have handling for list of base64 video frame

    # not part of openai api
    leading_prompt: Optional[str] = Field(default="", description="The string to append to the end of the prompt, this will not be part of the  generated response")
    prefix_strings: Optional[Union[str, List[str]]] = Field(default=None, description="String or list of strings to start the generation with. Can not be used together with regex_prefix_pattern")
    regex_pattern: Optional[constr(min_length=1)] = None   # regex to enforce
    regex_prefix_pattern: Optional[constr(min_length=1)] = Field(default=None, description="regex to enforce in the beginning of the generation, can not be used together with prefix_string")
    stop_words: Optional[List[str]] = Field(default=None, alias="stop")     # OpenAI use stop
    thinking_template: Optional[str] = None

    # tool call
    tool_call_thinking: bool = Field(default= True, description="Automatically trigger one liner tool call thinking when tool in auto mode to decide if tool is required")
    tool_call_thinking_token: int = Field(default= 200, description="Maximum token for tool thinking generation. If it exceed this threshold, no tool thinking is returned")
    tool_instruction_position: Literal["prefix", "postfix"] = (
        Field(default="prefix", description="Position of the general instruction to use tool. prefix for best kv caching"))
    tool_schema_position: Literal["prefix", "postfix"] = (
        Field(default="postfix", description="Position of the schema of individual tools. If tool_schema is unchanged through out, "
                                            "keep it as prefix for maximum kv caching. postfix for cases where tool are changing between api request"))

    artifact: Optional[Literal["No", "Fast", "Slow"]] = Field(default="No", description="Normal will parse the streamed output for artifact, whereas Strict is slower and will use format enforcer to enforce")
    return_thinking: Optional[Literal[False, True, "separate"]] = Field(
        default=False,
        description="Return the generated thinking to front end. False - not return, True - return, 'separate' - return separately as .thinking field. If used together with artifact, True will return as separate."
    )
    guided_decoding_backend: Optional[Literal["auto", "formatron", "lm-format-enforcer"]] = Field(
        default="auto",
        description="guided decoding backend. auto will choose the most suitable. If the selected backend is not working for specific llm backend (e.g. formatrong not working with llama cpp), selection will be auto"
    )

    # not yet supported options from here # TODO
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[Dict[int, float]] = {}
    top_logprob: int = None
    n: int = 1
    presence_penalty: float = 0
    response_format: Optional[ResponseFormat] = None
    seed: Optional[int] = None
    stream_options: Optional[StreamOption] = None
    parallel_tool_calls: bool = True            # default let the model handle and can not toggle

    @validator('regex_pattern', 'regex_prefix_pattern')
    def validate_regex(cls, v):
        """ this function is to handle special regex patterns """
        if v is not None:
            try:
                re.compile(v)
            except re.error as e:
                raise ValueError(f'Invalid regex pattern: {e}')
        return v

    @validator('tool_call_thinking_token')
    def validate_tool_call_thinking_token(cls, v):
        """ Validate that tool_call_thinking_token is greater than or equal to 0 """
        if v < 0:
            raise ValueError('tool_call_thinking_token must be greater than or equal to 0')
        return v



# from here on is answer model for response to api request
class OneTool(BaseModel):
    """The format to use to call one tool"""
    name: str = Field(description='name of the function to use')
    arguments: str


class ToolCallResponse(BaseModel):
    index: Optional[int] = None
    id: Optional[str] = None
    function: OneTool
    type: Optional[str] = "function"


class ToolCalling(BaseModel):
    """ The format to call one or multiple tools """
    functions_calling: List[OneTool] = Field(description='the list of functions to call in chronological order',
                                             default=[])


class ChatResponse(BaseModel):
    role: Literal['system', 'user', 'assistant'] = None
    tool_call_id: Optional[str] = None
    content: Optional[str] = None
    # name: Optional[str] = None
    # function_call: Optional[ToolCalling] = None   # depreciated
    tool_calls: Optional[List[ToolCallResponse]] = None


class ChatMessage(BaseModel):
    role: Literal['system', 'user', 'assistant'] = None
    tool_call_id: Optional[str] = None
    content: Optional[str] = None
    name: Optional[str] = None
    # thinking
    thinking: Optional[str] = None

    # function_call: Optional[ToolCalling] = None   # depreciated
    tool_calls: Optional[List[ToolCallResponse]] = None
    artifact_meta: Union[TextTag, ArtifactTag] = None

    def __str__(self) -> str:
        if self.role == "system":
            return f"system:\n{self.content}\n"

        elif self.role == "function":
            return f"function name={self.name}:\n{self.content}\n"

        elif self.role == "user":
            if self.content is None:
                return "user:\n</s>"
            else:
                return f"user:\n</s>{self.content}\n"

        elif self.role == "assistant":
            if self.content is not None and self.function_call is not None:
                return f"assistant:\n{self.content}\nassistant to={self.function_call.name}:\n{self.function_call.arguments}</s>"

            elif self.function_call is not None:
                return f"assistant to={self.function_call.name}:\n{self.function_call.arguments}</s>"

            elif self.content is None:
                return "assistant"

            else:
                return f"assistant:\n{self.content}\n"
        else:
            raise ValueError(f"Unsupported role: {self.role}")


class UsageResponse(BaseModel):
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0


class Choice(BaseModel):
    index: int = 0
    message: ChatMessage
    logprobs: Union[float, None] = None
    finish_reason: Optional[
        Literal["stop", "length", "tool_calls", "content_filter"]
    ] = "stop"

    @classmethod
    def from_message(cls, message: ChatMessage, finish_reason: str):
        return cls(message=message, finish_reason=finish_reason)


class ChoiceDeltaToolCallFunction(BaseModel):
    arguments: Optional[str] = None
    name: Optional[str] = None


class ChoiceDeltaToolCall(BaseModel):
    index: Optional[int] = None
    id: Optional[str] = None
    function: Optional[ChoiceDeltaToolCallFunction] = None
    type: Optional[str] = None


class ChoiceDelta(BaseModel):
    content: Optional[str] = None
    function_call: Optional[str] = None
    refusal: Optional[str] = None
    role: Optional[str] = None
    tool_calls: Optional[List[ChoiceDeltaToolCall]] = None


class StreamChoice(BaseModel):
    index: int
    delta: ChoiceDelta
    finish_reason: Optional[Literal["stop", "length", "tool_calls"]] = None


class ChatCompletionResponse(BaseModel):
    id: str = Field(description='id of the response', default_factory=lambda: "cmpl-" + str(uuid.uuid4().hex))
    object: Literal["chat.completion", "chat.completion.chunk"] = Field(default="chat.completion")
    created: int = Field(description='timestamp when object was created', default_factory=lambda: int(time.time()))
    model: str = Field(description='name of the model')
    choices: List[Union[Choice, StreamChoice]]
    usage: UsageResponse = None


class CompletionChoice(BaseModel):
    text: str
    index: int
    logprobs: Optional[dict] = None
    finish_reason: Optional[str] = None


class CompletionResponse(BaseModel):
    id: str = Field(default_factory=lambda: f"cmpl-{uuid.uuid4().hex}")
    object: Literal["text_completion"] = "text_completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    system_fingerprint: str = Field(default="fp_44709d6fcb")
    choices: List[CompletionChoice]
    usage: Optional[UsageResponse] = None


class CompletionStreamResponse(BaseModel):
    id: str = Field(default_factory=lambda: f"cmpl-{uuid.uuid4().hex}")
    object: Literal["text_completion"] = "text_completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    system_fingerprint: str = Field(default="fp_44709d6fcb")
    choices: List[CompletionChoice]


# embedding dataclass from here
class EmbeddingRequest(BaseModel):
    """ Request to embedding some text in the input"""
    input: Union[str, List[str], List[List[int]]] = Field(description="text to embed, a str or list of str")
    model: str
    dimension: Optional[int] = None
    encoding_format: Optional[Literal["float", "base64"]] = "float"


class EmbeddingObject(BaseModel):
    index: int = Field(description='index of the embedding', default=0)
    object: str = 'embedding'
    embedding: Union[List[float], str] = Field(description='list of float for embedding vector and str for base64')


class EmbeddingResponse(BaseModel):
    class Usage(BaseModel):
        prompt_tokens: int = Field(description='number of tokens in the prompt')
        total_tokens: int = Field(description='total number of tokens')

    object: str = 'list'
    model: str = Field(description='name of the model')
    usage: Usage
    data: List[EmbeddingObject]


class GenerateQuery(BaseModel):
    prompt: str = Field(description='prompt')
    model: Optional[str] = Field(description='name of the model', default=None)
    stream: Optional[bool] = False
    max_tokens: Optional[int] = Field(description='max number of tokens', default=None)


# class GenerationStats(BaseModel):
#     input_tokens_count: int = Field(description='input tokens count', default=0)
#     output_tokens_count: int = Field(description='output tokens count', default=0)
#     time_to_first_token: float = Field(description='time to first token', default=0)
#     time_generate: float = Field(description='time to generate tokens', default=0)
#
#     @property
#     def generation_speed(self) -> float:
#         if self.time_generate > 0:
#             return round(self.output_tokens_count / self.time_generate, ndigits=1)
#         else:
#             return 0
#
#     @property
#     def total_time(self) -> float:
#         return round(self.time_to_first_token + self.time_generate, ndigits=1)
#
#     @property
#     def prefill_speed(self) -> float:
#         if self.time_to_first_token > 0:
#             return round(self.input_tokens_count / self.time_to_first_token, ndigits=1)
#         else:
#             return 0

    @property
    def total_tokens_count(self) -> float:
        return self.input_tokens_count + self.output_tokens_count


class ModelObject(BaseModel):
    id: str = Field(description='id of the model')
    object: str = Field(description='object type', default="model")
    owned_by: str = Field(description='object owner', default="remichu")
    created_by: int = Field(description='model creation time', default=1686935002)


class ModelObjectResponse(BaseModel):
    object: str = Field(description='object type', default="list")
    data: List[ModelObject] = []


class Thinking(BaseModel):
    xml: str = Field(description='xml string')
    regex: str = Field(description='regex string to enforce any value', default=None)


class VoiceConfig(BaseModel):
    """Configuration for a single voice sample"""
    ref_audio_path: str = Field(description='path to the sound sample')
    ref_audio_transcription: str = Field(description='text of the voice sample')
    language: str = Field(description='language of the voice sample')
    speed_factor: Optional[float] = Field(description='speed factor of the voice sample', default=1.0)

    model_config = ConfigDict(extra="forbid")



# list of supported backend. None meaning it the api will take backend set from yaml config file
SUPPORTED_BACKENDS = [
    "exllama",
    "llama_cpp",
    "transformers",
    "mlx_vlm",
    "embedding",
    "faster_whisper",
    "mlx_whisper",
    "gpt_sovits",
    "kokoro",
    None
]

class ModelSpec(BaseModel):
    model_id: Optional[str] = Field(description='id of the model which should be the path to the model', default=None)
    model_name: Optional[str] = Field(description='name of the model, which is the key inside yml configuration file', default=None)
    model_type: Optional[Literal["stt", "llm", "tts", "embedding", None]] = Field(description='type of the model, will be automatically determined based on backend', default=None)
    gpus: Optional[Union[Literal["auto"], List[float]]] = Field(description='VRam usage for each GPU', default="auto")
    cache_size: Optional[int] = Field(default=None, description='The context length for cache text in int. If None, will be set to the model context length')
    cache_quant: Optional[Literal["FP16", "Q4", "Q6", "Q8"]] = Field(default=None, description='the quantization to use for cache, will use Q4 if not specified')
    max_seq_len: Optional[int] = Field(description="max sequence length", default=None)
    backend: Optional[Union[Literal[tuple(SUPPORTED_BACKENDS)], None]] = Field(description="model engine backend", default=None)
    tensor_parallel: Optional[bool] = Field(description="tensor parallel mode", default=False)
    prompt_template: Optional[str] = Field(description="prompt template", default=None)
    eos_token_list: List[str] = Field(description="eos tokens, can customize token here", default_factory=list)

    quant: Optional[float] = Field(description="quantization if the model support quantization on the fly", default=None)

    # number of concurrent request this model can handle
    max_concurrent_requests: int = Field(description="number of concurrent request this model can handle", default=1)

    # extra argument for specific backend or model
    backend_extra_args: Optional[Dict[Any, Any]] = Field(description="extra args to pass to the backend", default_factory=dict)

    # speculative decoding
    draft_model_id: Optional[str] = Field(description='id of the draft model', default=None)
    draft_model_name: Optional[str] = Field(description='name of the draft model', default=None)
    draft_gpus: Optional[Union[Literal["auto"], List[float]]] = Field(description='VRam usage for each GPU', default="auto")
    draft_cache_size: Optional[int] = Field(description='The context length for cache text in int. If None, will be set to the model context length', default=None)
    draft_cache_quant: Optional[Literal["FP16", "Q4", "Q6", "Q8"]] = Field(default=None, description='the quantization to use for cache, will use Q4 if not specified')
    # backend is assumed to be the same as main model


    # whether require api call to match model name or not
    strict: bool = Field(description="whether require api call to match model name or not", default=False)

    # argument for different voice sample
    voice: Optional[Dict[str, VoiceConfig]] = Field(
        default=None,
        description="Voice configurations mapping voice names to their settings"
    )

    # audio related setting
    language: Optional[str] = Field(description="language of the audio", default="auto")

    # dont allow non recognizable option
    model_config = ConfigDict(extra="forbid", validate_assignment=True, protected_namespaces=())  # disable protected_namespaces due to it field use model_ in the name


    @validator('gpus', pre=True, always=True)
    def validate_gpus(cls, v):
        if v is None:
            return None
        if isinstance(v, dict):
            # Convert the dict to a list based on GPU IDs
            return [v.get(i, 0.0) for i in range(torch.cuda.device_count())]
        return v

    # TODO this is clasing with embedding cause embedding will set visiable GPU and hence it is not seen anymore in this validator
    # @validator('gpus')
    # def check_gpus(cls, gpus):
    #     if gpus is None:
    #         return None
    #     num_gpus = torch.cuda.device_count()
    #     for gpu_id, vram in enumerate(gpus):
    #         if gpu_id < 0 or gpu_id >= num_gpus:
    #             raise ValueError(f"Invalid GPU ID {gpu_id}. Must be between 0 and {num_gpus - 1}")
    #
    #         if vram < 0:
    #             raise ValueError(f"VRAM usage for GPU {gpu_id} must be a non-negative number")
    #
    #         if vram > 0:
    #             device = torch.cuda.get_device_properties(gpu_id)
    #             total_vram = device.total_memory / (1024 ** 3)  # Convert bytes to GB
    #             if vram > total_vram:
    #                 raise ValueError(
    #                     f"Requested VRAM ({vram} GB) for GPU {gpu_id} exceeds available VRAM ({total_vram:.2f} GB)")

        # return gpus


    @classmethod
    def get_model_type_from_backend(cls, backend: str = None):
        if backend is None:
            return None
        elif backend in ["exllama", "llama_cpp", "transformers", "mlx_vlm"]:
            return "llm"
        elif backend in ["faster_whisper", "mlx_whisper"]:
            return "stt"
        elif backend in ["gpt_sovits", "kokoro"]:
            return "tts"
        elif backend in ["embedding"]:
            return "embedding"

    @classmethod
    def from_dict(cls, input_data: Union[str, Dict[str, Any]]):
        if isinstance(input_data, str):
            # If input is a string, split it into a dictionary
            params = input_data.split()
            input_dict = {}
            for param in params:
                key, value = param.split('=')
                input_dict[key] = value.strip("'")  # Strip single quotes here as well
        else:
            # If input is already a dictionary, use it as is
            input_dict = input_data

        model_id = input_dict.get('model_id')
        if model_id:
            model_id = input_dict.get('model_id').strip("'")  # Remove single quotes if present
        model_name = input_dict.get('model_name')
        max_seq_len = input_dict.get('max_seq_len', None)
        gpus = input_dict.get('gpus')
        cache_size = input_dict.get('cache_size')
        backend = input_dict.get('backend', None)  # Default to None if not provided
        tensor_parallel = input_dict.get('tp', False)
        if tensor_parallel=="True" or tensor_parallel=="true":
            tensor_parallel = True

        # TODO clean up code and merge config setting into config manager
        if backend == "None":
            backend = None
        cache_quant = input_dict.get('cache_quant', None)

        if gpus and isinstance(gpus, str) and gpus.lower() != "auto":
            gpus = [float(x) for x in gpus.split(',')]

        if cache_size:
            cache_size = int(cache_size)

        if cache_size:
            cache_size = int(cache_size)

        strict = input_dict.get('strict', False)
        voice = input_dict.get('voice', {})

        model_type = input_dict.get('model_type', None)
        if model_type is None:
            model_type = cls.get_model_type_from_backend(backend)

        prompt_template = input_dict.get('prompt_template', None)

        # concurrent request
        allowed_concurrency = 50 if backend in ["exllama", "embedding"] else 1  # TODO to look into optimal number for each backend
        max_concurrent_requests = input_dict.get('max_concurrent_requests', allowed_concurrency)

        backend_extra_args = input_dict.get('backend_extra_args', {})

        # speculative decoding
        draft_model_id = input_dict.get('draft_model_id')
        if draft_model_id:
            draft_model_id = draft_model_id.strip("'")  # Remove single quotes if present

        draft_model_name = input_dict.get('draft_model_name', None)

        draft_gpus = input_dict.get('draft_gpus')
        draft_cache_size = input_dict.get('draft_cache_size')
        draft_cache_quant = input_dict.get('draft_cache_quant', None)

        if draft_gpus and isinstance(draft_gpus, str) and draft_gpus.lower() != "auto":
            draft_gpus = [float(x) for x in draft_gpus.split(',')]

        if draft_cache_size:
            draft_cache_size = int(draft_cache_size)

        return cls(model_id=model_id, model_name=model_name, model_type=model_type,
                   gpus=gpus, cache_size=cache_size, backend=backend, cache_quant=cache_quant,
                   strict=strict,
                   voice=voice,
                   prompt_template=prompt_template,
                   backend_extra_args=backend_extra_args,
                   max_concurrent_requests=max_concurrent_requests,
                   max_seq_len=max_seq_len,
                   tensor_parallel=tensor_parallel,
                   draft_model_id=draft_model_id, draft_model_name=draft_model_name,
                   draft_gpus=draft_gpus, draft_cache_size=draft_cache_size, draft_cache_quant=draft_cache_quant)

    def get_visible_gpu_indices(self) -> str:
        """
        Generate a string of GPU indices based on allocated GPUs.
        If no GPUs are specified, return all available GPU indices.

        Returns:
            str: A comma-separated string of GPU indices with allocated VRAM,
                 or all available GPU indices if none are specified.
        """
        if self.gpus is None or self.gpus == "auto":
            import torch
            return ','.join(str(i) for i in range(torch.cuda.device_count()))

        if all(vram == 0 for vram in self.gpus):
            return ""  # No GPUs allocated

        visible_devices = [str(i) for i, vram in enumerate(self.gpus) if vram > 0]
        return ','.join(visible_devices)

    @staticmethod
    def deep_merge_dicts(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> Dict[str, Any]:
        """
        Recursively merge two dictionaries, handling nested dictionaries.
        Values from dict2 take precedence over dict1 except for nested dictionaries,
        which are merged recursively.

        Args:
            dict1 (Dict[str, Any]): First dictionary
            dict2 (Dict[str, Any]): Second dictionary (takes precedence for non-dict values)

        Returns:
            Dict[str, Any]: Merged dictionary
        """
        merged = dict1.copy()

        for key, value in dict2.items():
            if key in merged and isinstance(merged[key], dict) and isinstance(value, dict):
                # If both values are dictionaries, merge them recursively
                merged[key] = ModelSpec.deep_merge_dicts(merged[key], value)
            else:
                # For non-dict values or when key doesn't exist in dict1,
                # take the value from dict2
                merged[key] = value

        return merged

    def merge_with_config(self, model_config: Dict[str, Any]) -> 'ModelSpec':
        """
        Merges the current ModelSpec instance with a model configuration dictionary.
        Values from the current ModelSpec take precedence over the model_config.
        Handles nested dictionaries by merging them recursively.

        Args:
            model_config (Dict[str, Any]): Configuration dictionary from config manager

        Returns:
            ModelSpec: A new ModelSpec instance with merged configurations
        """
        # Convert current ModelSpec to dict, excluding None values
        spec_dict = self.model_dump(exclude_none=True)

        # Merge configurations recursively
        merged_config = ModelSpec.deep_merge_dicts(model_config, spec_dict)

        # Create new ModelSpec instance with merged configuration
        return ModelSpec(**merged_config)

    @classmethod
    def from_merged_config(cls, model_spec: 'ModelSpec', model_config: Dict[str, Any]) -> 'ModelSpec':
        """
        Class method to create a new ModelSpec instance by merging an existing ModelSpec
        with a model configuration dictionary.

        Args:
            model_spec (ModelSpec): Existing ModelSpec instance
            model_config (Dict[str, Any]): Configuration dictionary from config manager

        Returns:
            ModelSpec: A new ModelSpec instance with merged configurations
        """
        return model_spec.merge_with_config(model_config)



class ModelDownloadSpec(BaseModel):
    """ dataclass for model download"""
    model_name: str
    quant: Optional[float] = None
    backend: Literal[tuple(SUPPORTED_BACKENDS)] = None

    # disable protected_namespaces due to it field use model_ in the name
    model_config = ConfigDict(extra="forbid", protected_namespaces=())



from fastapi import WebSocket, WebSocketDisconnect, APIRouter, Query
import logging
from gallama.realtime.websocket_handler import WebSocketMessageHandler
from ..data_classes import ModelInstanceInfo
from ..realtime.websocket_manager import WebSocketManager
from ..realtime.session_manager import SessionManager
from ..dependencies_server import get_server_manager, get_server_logger
import websockets

# Create router
router = APIRouter(prefix="", tags=["realtime"])

server_manager = get_server_manager()

logger = get_server_logger()

# Initialize managers
session_manager = SessionManager()
message_handler = WebSocketMessageHandler()
#     stt_url="ws://localhost:8001/speech-to-text",
#     llm_url="ws://localhost:8002/llm",
#     tts_url="ws://localhost:8003/ws/speech"
# )
websocket_manager = WebSocketManager(session_manager, message_handler)

@router.websocket("/video")
async def websocket_video(websocket: WebSocket):
    await websocket.accept()
    llm_websocket = None  # Initialize LLM WebSocket connection
    try:
        # Get the LLM instance and its WebSocket URL
        llm_instance = server_manager.get_instance(model_type="llm")
        llm_websocket_url = f"ws://localhost:{llm_instance.port}/video"

        # Connect directly to the LLM WebSocket
        llm_websocket = await websockets.connect(llm_websocket_url)
        logger.debug(f"Connected to LLM WebSocket at {llm_websocket_url}")

        while True:
            # Receive raw data from the client
            data = await websocket.receive_bytes()
            logger.debug(f"Received {len(data)} bytes from /video WebSocket")

            # Forward the raw data directly to the LLM WebSocket
            await llm_websocket.send(data)

            # Optional: Send an acknowledgment back to the client
            await websocket.send_text("Data forwarded to LLM WebSocket")

    except WebSocketDisconnect:
        logger.info("Client disconnected from /video WebSocket")
    except Exception as e:
        logger.error(f"Error in /video WebSocket: {str(e)}")
    finally:
        # Close the LLM WebSocket connection if it exists
        if llm_websocket:
            await llm_websocket.close()
        # Close the /video WebSocket connection gracefully
        try:
            await websocket.close()
        except RuntimeError as e:
            logger.debug(f"Websocket already closed: {str(e)}")


@router.websocket("/{path:path}")
async def websocket_endpoint(
    websocket: WebSocket,
    path: str,
    model: str = Query(..., description="Realtime model ID to connect to"),
):
    """
    WebSocket endpoint matching OpenAI's Realtime API specifications

    Connection URL: wss://api.openai.com/v1/realtime
    Required query parameters:
    - model: Realtime model ID (e.g., gpt-4o-realtime-preview-2024-12-17)
    Required headers:
    - Authorization: Bearer YOUR_API_KEY
    - OpenAI-Beta: realtime=v1
    """
    try:
        # Get headers
        headers = dict(websocket.headers)
        authorization = headers.get("authorization", "")
        openai_beta = headers.get("openai-beta")

        # Accept WebSocket connection with OpenAI protocol
        # protocols = websocket.headers.get("sec-websocket-protocol", "").split(", ")
        # if "openai-beta.realtime-v1" in protocols:
        #     await websocket.accept(subprotocol="openai-beta.realtime-v1")
        # else:
        #     await websocket.accept()

        api_key = None
        if authorization and authorization.startswith("Bearer "):
            api_key = authorization.replace("Bearer ", "")

        stt_instance = server_manager.get_instance(model_type="stt")
        llm_instance = server_manager.get_instance(model_type="llm")
        tts_instance = server_manager.get_instance(model_type="tts")

        # Create session
        session = await websocket_manager.initialize_session(
            websocket,
            model=model,
            api_key=api_key,
            stt_url=f"ws://localhost:{stt_instance.port}/speech-to-text",
            llm_url=f"ws://localhost:{llm_instance.port}/llm",
            tts_url=f"ws://localhost:{tts_instance.port}/ws/speech",
        )

        try:
            await websocket_manager.start_background_tasks(session, websocket)

            while True:
                message = await websocket.receive_json()
                await message_handler.handle_message(websocket, session, message)

        except WebSocketDisconnect:
            await websocket_manager.stop_background_tasks(session)
            await message_handler.cleanup()
            await session_manager.delete_session(session.id)
            logging.info(f"WebSocket disconnected for session {session.id}")

        except Exception as e:
            logging.error(f"Error in websocket connection: {str(e)}")
            await websocket.close(code=4003, reason="Internal server error")

    except Exception as e:
        logging.error(f"Error establishing websocket connection: {str(e)}")
        await websocket.close(code=4003, reason="Connection error")





from ..base import ModelInterface
from typing import Optional, Dict, List, Union
from fastapi import Request                 # for type hint
import time

# for async running of llama cpp
import asyncio
from concurrent.futures import ThreadPoolExecutor

try:
    from llama_cpp import Llama
    from llama_cpp import LogitsProcessorList
except:
    Llama = None
    LogitsProcessorList = None

# format enforcement with formatron
try:
    from formatron.formatter import FormatterBuilder
except ImportError:
    FormatterBuilder = None

# format enforcement with llama cpp
try:
    from lmformatenforcer.tokenenforcer import TokenEnforcerTokenizerData
    from lmformatenforcer.integrations.llamacpp import (
        build_llamacpp_logits_processor,
        build_token_enforcer_tokenizer_data as build_token_enforcer_tokenizer_data_llama_cpp
    )
except:
    # llama_cpp optional dependency
    build_llamacpp_logits_processor = None
    build_token_enforcer_tokenizer_data_llama_cpp = None
    LogitsProcessorList = None

# custom data classes
from gallama.data_classes import (
    ModelSpec,
    GenStart,
    GenEnd,
    GenQueue,
    GenText,
    GenerationStats,
    QueueContext,
    GenQueueDynamic
)
from gallama.logger.logger import logger


class ModelLlamaCpp(ModelInterface):
    def __init__(self, model_spec:ModelSpec):
        super().__init__(model_spec)

        self.model, self.tokenizer = self.load_model()

        # initialize lmfe tokenizer data
        self.lm_enforcer_tokenizer_data = build_token_enforcer_tokenizer_data_llama_cpp(self.model)


    def load_model(self):
        """Load the model, tokenizer, cache, and optional processor."""

        model, tokenizer = self.load_model_llama_cpp(
            model_id=self.model_id,
            gpus=self.gpus,
        )

        # load draft model
        if self.draft_model_id is not None:
            raise "Draft model currently not supported for llama cpp backend"

        self.eos_token_ids = self.generate_eos_tokens_id()
        return model, tokenizer


    def load_model_llama_cpp(self, model_id, gpus):
        """This function return the model and its tokenizer"""
        logger.info("Loading model: " + model_id)

        # currently speculative decoding not supported by model specific for LLama CPP python
        if isinstance(gpus, str) and gpus == "auto":
            model = Llama(
                model_path=self.model_id,
                n_gpu_layers=-1,
                seed=1,
                n_ctx=self.max_seq_len if self.max_seq_len else 0,  # Uncomment to increase the context window
                flash_attn=True,
                offload_kqv=True,
                # draft_model=draf_model_id,
            )
        elif isinstance(gpus, list):  # user specify the gpus split
            model = Llama(
                model_path=self.model_id,
                n_gpu_layers=-1,
                seed=1,
                n_ctx=self.max_seq_len if self.max_seq_len else 0,  # Uncomment to increase the context window
                flash_attn=True,
                offload_kqv=True,
                tensor_split=gpus,
                # draf_model_id=draf_model_id,
            )
        else:
            raise ValueError("Device map should be either 'auto', 'gpu' split")

        # set max_seq_len based on model
        self.max_seq_len = model._model.n_ctx_train()
        tokenizer = model       # llama cpp doesnt have separate tokenizer object

        self.eos_token_ids = self.generate_eos_tokens_id()

        return model, tokenizer



    def generate_eos_tokens_id(self) -> List[int]:
        """Generate the end-of-sequence token IDs."""

        # currently the stop word for llama cpp work by string and not by token id
        # hence usage of this is not required
        return []


    # ************* method for generation from here
    def _run_generator_and_queue(self, prompt, logits_processor, max_tokens, temperature, stop, gen_queue_list,
                                 top_p=0.8, prefix_strings=None, stop_word_to_return="", gen_type_str: str="text"):

        loop = asyncio.get_event_loop()

        generator = self.model(
            prompt=prompt,
            #suffix=prefix_strings,
            logits_processor=logits_processor,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            min_p=0.05,
            stop=stop,
            repeat_penalty=1.1,
            stream=True,
        )

        generate_text = ""
        for chunk in generator:
            chunk_text = GenText(content=chunk['choices'][0]['text'], text_type=gen_type_str)
            generate_text += chunk_text.content
            for g_queue in gen_queue_list:
                # g_queue.get_queue().put_nowait(chunk_text)
                future = asyncio.run_coroutine_threadsafe(
                    g_queue.put(chunk_text),
                    loop
                )
                future.result()

        # return stop_word if there is
        if stop_word_to_return:
            chunk_text = GenText(content=stop_word_to_return, text_type=gen_type_str)
            generate_text += chunk_text.content
            for g_queue in gen_queue_list:
                # g_queue.get_queue().put_nowait(chunk_text)
                future = asyncio.run_coroutine_threadsafe(
                    g_queue.put(chunk_text),
                    loop
                )
                future.result()
        return generate_text

    async def generate(
        self,
        prompt: str,
        gen_queue: Union[GenQueue, QueueContext, List[QueueContext]],
        request: Optional[Request] = None,          # for disconnection check, however, no stopping feature for llama cpp
        gen_type: Union[str, GenStart] = "text",    # the generated result will be store to this queue
        temperature: float = 0.01,
        top_p: float = 0.8,
        formatter: FormatterBuilder | TokenEnforcerTokenizerData = None,
        stop_words: Union[List[str], str] = None,
        prefix_strings: Optional[Union[str, List[str]]] = None,
        banned_strings: list[str] | None = None,
        max_tokens: int = None,
        quiet=False,
        messages: List = None,  # query.message for multimodal
        **kwargs,
    ) -> (str, GenerationStats):

        loop = asyncio.get_event_loop()

        if not quiet:
            logger.info("----------------------Prompt---------------\n" + prompt)
            logger.debug("----------------------temperature---------\n" + str(temperature))

        # ensure that generator is initialized
        # if self.pipeline is None:
        #     self.pipeline = await self._get_pipeline_async()

        # make gen_queue to List[QueueContext] for standardize downstream handline
        # Convert gen_queue to List[GenQueueDynamic] format to standardize downstream handling
        gen_queue_list = []
        if isinstance(gen_queue, GenQueueDynamic):
            gen_queue_list = [gen_queue]
        elif isinstance(gen_queue, GenQueue):
            # Wrap the GenQueue in a GenQueueDynamic
            gen_queue_list = [GenQueueDynamic(existing_queue=gen_queue, include_GenStats=True, include_GenEnd=True)]
        elif isinstance(gen_queue, list):
            # Ensure all items in the list are GenQueueDynamic objects
            for queue in gen_queue:
                if isinstance(queue, GenQueueDynamic):
                    gen_queue_list.append(queue)
                elif isinstance(queue, GenQueue):
                    # Wrap the GenQueue in a GenQueueDynamic
                    gen_queue_list.append(
                        GenQueueDynamic(existing_queue=queue, include_GenStats=True, include_GenEnd=True))
                else:
                    raise TypeError("gen_queue list must contain only GenQueue or GenQueueDynamic objects")
        else:
            raise TypeError("gen_queue must be either a GenQueue, GenQueueDynamic, or a list of GenQueueDynamic")

        # convert prompt to token id
        input_ids = self.tokenizer.tokenize(prompt.encode("utf-8"), add_bos=False)
        self.validate_token_length(len(input_ids))

        # format enforcer
        logits_processors = None
        if formatter:
            logits_processors = LogitsProcessorList([
                build_llamacpp_logits_processor(
                    llm=self.pipeline.lm_enforcer_tokenizer_data,
                    character_level_parser=formatter,
                )
            ])

        start_time = time.time()

        # find stop conditions
        stop_word_to_return = ""
        if stop_words:
            if isinstance(stop_words, str):
                stop_word_to_return = stop_words
                stop_words = [stop_words]

            elif isinstance(stop_words, list):
                stop_word_to_return = stop_words[0]

            if not self.eos_token_str:
                raise Exception("EOS token not set in model_config")
            stop_conditions = self.eos_token_str + stop_words  # concat the 2 list
            logger.debug("stop_words: " + str(stop_conditions))
        else:
            stop_conditions = self.eos_token_str

        max_tokens_to_use = min(
            self.max_seq_len - len(input_ids),
            max_tokens, 4096) if max_tokens else min(self.max_seq_len - len(input_ids),
                                                     4096
                                                     )

        # kick-start the generation and let down stream know gen type
        if isinstance(gen_type, str):
            gen_type_str = gen_type
            gen_type = GenStart(gen_type=gen_type)
        else:
            gen_type_str = gen_type.gen_type  # get out the generation type in str format

        for g_queue in gen_queue_list:
            # g_queue.get_queue().put_nowait(gen_type)
            future = asyncio.run_coroutine_threadsafe(
                g_queue.put(gen_type),
                loop
            )
            future.result()

        # Create a task to check for disconnection
        # llama cpp python generator is not async, hence running fake async..
        # Run the synchronous generator in a separate thread
        with ThreadPoolExecutor() as pool:
            generate_text = await loop.run_in_executor(
                pool,
                self._run_generator_and_queue,
                prompt, logits_processors, max_tokens_to_use, temperature, stop_conditions, gen_queue_list,
                top_p, prefix_strings, stop_word_to_return, gen_type_str
            )

        duration = time.time() - start_time

        gen_stats = GenerationStats()
        for g_queue in gen_queue_list:
            if g_queue.include_GenStats:
                # g_queue.get_queue().put_nowait(gen_stats)
                future = asyncio.run_coroutine_threadsafe(
                    g_queue.put(gen_stats),
                    loop
                )
                future.result()


        # this to signal the end of generation
        for g_queue in gen_queue_list:
            if g_queue.include_GenEnd:
                # g_queue.get_queue().put_nowait(GenEnd())
                future = asyncio.run_coroutine_threadsafe(
                    g_queue.put(GenEnd()),
                    loop
                )
                future.result()
        logger.debug("----------------------LLM Raw Response---------------\n" + generate_text)


import os
import random

import numpy as np
import torch
import torch.utils.data
from tqdm import tqdm

from . import spec_utils


class VocalRemoverValidationSet(torch.utils.data.Dataset):
    def __init__(self, patch_list):
        self.patch_list = patch_list

    def __len__(self):
        return len(self.patch_list)

    def __getitem__(self, idx):
        path = self.patch_list[idx]
        data = np.load(path)

        X, y = data["X"], data["y"]

        X_mag = np.abs(X)
        y_mag = np.abs(y)

        return X_mag, y_mag


def make_pair(mix_dir, inst_dir):
    input_exts = [".wav", ".m4a", ".mp3", ".mp4", ".flac"]

    X_list = sorted(
        [
            os.path.join(mix_dir, fname)
            for fname in os.listdir(mix_dir)
            if os.path.splitext(fname)[1] in input_exts
        ]
    )
    y_list = sorted(
        [
            os.path.join(inst_dir, fname)
            for fname in os.listdir(inst_dir)
            if os.path.splitext(fname)[1] in input_exts
        ]
    )

    filelist = list(zip(X_list, y_list))

    return filelist


def train_val_split(dataset_dir, split_mode, val_rate, val_filelist):
    if split_mode == "random":
        filelist = make_pair(
            os.path.join(dataset_dir, "mixtures"),
            os.path.join(dataset_dir, "instruments"),
        )

        random.shuffle(filelist)

        if len(val_filelist) == 0:
            val_size = int(len(filelist) * val_rate)
            train_filelist = filelist[:-val_size]
            val_filelist = filelist[-val_size:]
        else:
            train_filelist = [
                pair for pair in filelist if list(pair) not in val_filelist
            ]
    elif split_mode == "subdirs":
        if len(val_filelist) != 0:
            raise ValueError(
                "The `val_filelist` option is not available in `subdirs` mode"
            )

        train_filelist = make_pair(
            os.path.join(dataset_dir, "training/mixtures"),
            os.path.join(dataset_dir, "training/instruments"),
        )

        val_filelist = make_pair(
            os.path.join(dataset_dir, "validation/mixtures"),
            os.path.join(dataset_dir, "validation/instruments"),
        )

    return train_filelist, val_filelist


def augment(X, y, reduction_rate, reduction_mask, mixup_rate, mixup_alpha):
    perm = np.random.permutation(len(X))
    for i, idx in enumerate(tqdm(perm)):
        if np.random.uniform() < reduction_rate:
            y[idx] = spec_utils.reduce_vocal_aggressively(
                X[idx], y[idx], reduction_mask
            )

        if np.random.uniform() < 0.5:
            # swap channel
            X[idx] = X[idx, ::-1]
            y[idx] = y[idx, ::-1]
        if np.random.uniform() < 0.02:
            # mono
            X[idx] = X[idx].mean(axis=0, keepdims=True)
            y[idx] = y[idx].mean(axis=0, keepdims=True)
        if np.random.uniform() < 0.02:
            # inst
            X[idx] = y[idx]

        if np.random.uniform() < mixup_rate and i < len(perm) - 1:
            lam = np.random.beta(mixup_alpha, mixup_alpha)
            X[idx] = lam * X[idx] + (1 - lam) * X[perm[i + 1]]
            y[idx] = lam * y[idx] + (1 - lam) * y[perm[i + 1]]

    return X, y


def make_padding(width, cropsize, offset):
    left = offset
    roi_size = cropsize - left * 2
    if roi_size == 0:
        roi_size = cropsize
    right = roi_size - (width % roi_size) + left

    return left, right, roi_size


def make_training_set(filelist, cropsize, patches, sr, hop_length, n_fft, offset):
    len_dataset = patches * len(filelist)

    X_dataset = np.zeros((len_dataset, 2, n_fft // 2 + 1, cropsize), dtype=np.complex64)
    y_dataset = np.zeros((len_dataset, 2, n_fft // 2 + 1, cropsize), dtype=np.complex64)

    for i, (X_path, y_path) in enumerate(tqdm(filelist)):
        X, y = spec_utils.cache_or_load(X_path, y_path, sr, hop_length, n_fft)
        coef = np.max([np.abs(X).max(), np.abs(y).max()])
        X, y = X / coef, y / coef

        l, r, roi_size = make_padding(X.shape[2], cropsize, offset)
        X_pad = np.pad(X, ((0, 0), (0, 0), (l, r)), mode="constant")
        y_pad = np.pad(y, ((0, 0), (0, 0), (l, r)), mode="constant")

        starts = np.random.randint(0, X_pad.shape[2] - cropsize, patches)
        ends = starts + cropsize
        for j in range(patches):
            idx = i * patches + j
            X_dataset[idx] = X_pad[:, :, starts[j] : ends[j]]
            y_dataset[idx] = y_pad[:, :, starts[j] : ends[j]]

    return X_dataset, y_dataset


def make_validation_set(filelist, cropsize, sr, hop_length, n_fft, offset):
    patch_list = []
    patch_dir = "cs{}_sr{}_hl{}_nf{}_of{}".format(
        cropsize, sr, hop_length, n_fft, offset
    )
    os.makedirs(patch_dir, exist_ok=True)

    for i, (X_path, y_path) in enumerate(tqdm(filelist)):
        basename = os.path.splitext(os.path.basename(X_path))[0]

        X, y = spec_utils.cache_or_load(X_path, y_path, sr, hop_length, n_fft)
        coef = np.max([np.abs(X).max(), np.abs(y).max()])
        X, y = X / coef, y / coef

        l, r, roi_size = make_padding(X.shape[2], cropsize, offset)
        X_pad = np.pad(X, ((0, 0), (0, 0), (l, r)), mode="constant")
        y_pad = np.pad(y, ((0, 0), (0, 0), (l, r)), mode="constant")

        len_dataset = int(np.ceil(X.shape[2] / roi_size))
        for j in range(len_dataset):
            outpath = os.path.join(patch_dir, "{}_p{}.npz".format(basename, j))
            start = j * roi_size
            if not os.path.exists(outpath):
                np.savez(
                    outpath,
                    X=X_pad[:, :, start : start + cropsize],
                    y=y_pad[:, :, start : start + cropsize],
                )
            patch_list.append(outpath)

    return VocalRemoverValidationSet(patch_list)



import torch
import torch.nn.functional as F
from torch import nn

from . import layers_new


class BaseNet(nn.Module):
    def __init__(
        self, nin, nout, nin_lstm, nout_lstm, dilations=((4, 2), (8, 4), (12, 6))
    ):
        super(BaseNet, self).__init__()
        self.enc1 = layers_new.Conv2DBNActiv(nin, nout, 3, 1, 1)
        self.enc2 = layers_new.Encoder(nout, nout * 2, 3, 2, 1)
        self.enc3 = layers_new.Encoder(nout * 2, nout * 4, 3, 2, 1)
        self.enc4 = layers_new.Encoder(nout * 4, nout * 6, 3, 2, 1)
        self.enc5 = layers_new.Encoder(nout * 6, nout * 8, 3, 2, 1)

        self.aspp = layers_new.ASPPModule(nout * 8, nout * 8, dilations, dropout=True)

        self.dec4 = layers_new.Decoder(nout * (6 + 8), nout * 6, 3, 1, 1)
        self.dec3 = layers_new.Decoder(nout * (4 + 6), nout * 4, 3, 1, 1)
        self.dec2 = layers_new.Decoder(nout * (2 + 4), nout * 2, 3, 1, 1)
        self.lstm_dec2 = layers_new.LSTMModule(nout * 2, nin_lstm, nout_lstm)
        self.dec1 = layers_new.Decoder(nout * (1 + 2) + 1, nout * 1, 3, 1, 1)

    def __call__(self, x):
        e1 = self.enc1(x)
        e2 = self.enc2(e1)
        e3 = self.enc3(e2)
        e4 = self.enc4(e3)
        e5 = self.enc5(e4)

        h = self.aspp(e5)

        h = self.dec4(h, e4)
        h = self.dec3(h, e3)
        h = self.dec2(h, e2)
        h = torch.cat([h, self.lstm_dec2(h)], dim=1)
        h = self.dec1(h, e1)

        return h


class CascadedNet(nn.Module):
    def __init__(self, n_fft, nout=32, nout_lstm=128):
        super(CascadedNet, self).__init__()

        self.max_bin = n_fft // 2
        self.output_bin = n_fft // 2 + 1
        self.nin_lstm = self.max_bin // 2
        self.offset = 64

        self.stg1_low_band_net = nn.Sequential(
            BaseNet(2, nout // 2, self.nin_lstm // 2, nout_lstm),
            layers_new.Conv2DBNActiv(nout // 2, nout // 4, 1, 1, 0),
        )

        self.stg1_high_band_net = BaseNet(
            2, nout // 4, self.nin_lstm // 2, nout_lstm // 2
        )

        self.stg2_low_band_net = nn.Sequential(
            BaseNet(nout // 4 + 2, nout, self.nin_lstm // 2, nout_lstm),
            layers_new.Conv2DBNActiv(nout, nout // 2, 1, 1, 0),
        )
        self.stg2_high_band_net = BaseNet(
            nout // 4 + 2, nout // 2, self.nin_lstm // 2, nout_lstm // 2
        )

        self.stg3_full_band_net = BaseNet(
            3 * nout // 4 + 2, nout, self.nin_lstm, nout_lstm
        )

        self.out = nn.Conv2d(nout, 2, 1, bias=False)
        self.aux_out = nn.Conv2d(3 * nout // 4, 2, 1, bias=False)

    def forward(self, x):
        x = x[:, :, : self.max_bin]

        bandw = x.size()[2] // 2
        l1_in = x[:, :, :bandw]
        h1_in = x[:, :, bandw:]
        l1 = self.stg1_low_band_net(l1_in)
        h1 = self.stg1_high_band_net(h1_in)
        aux1 = torch.cat([l1, h1], dim=2)

        l2_in = torch.cat([l1_in, l1], dim=1)
        h2_in = torch.cat([h1_in, h1], dim=1)
        l2 = self.stg2_low_band_net(l2_in)
        h2 = self.stg2_high_band_net(h2_in)
        aux2 = torch.cat([l2, h2], dim=2)

        f3_in = torch.cat([x, aux1, aux2], dim=1)
        f3 = self.stg3_full_band_net(f3_in)

        mask = torch.sigmoid(self.out(f3))
        mask = F.pad(
            input=mask,
            pad=(0, 0, 0, self.output_bin - mask.size()[2]),
            mode="replicate",
        )

        if self.training:
            aux = torch.cat([aux1, aux2], dim=1)
            aux = torch.sigmoid(self.aux_out(aux))
            aux = F.pad(
                input=aux,
                pad=(0, 0, 0, self.output_bin - aux.size()[2]),
                mode="replicate",
            )
            return mask, aux
        else:
            return mask

    def predict_mask(self, x):
        mask = self.forward(x)

        if self.offset > 0:
            mask = mask[:, :, :, self.offset : -self.offset]
            assert mask.size()[3] > 0

        return mask

    def predict(self, x, aggressiveness=None):
        mask = self.forward(x)
        pred_mag = x * mask

        if self.offset > 0:
            pred_mag = pred_mag[:, :, :, self.offset : -self.offset]
            assert pred_mag.size()[3] > 0

        return pred_mag



from ..data_classes import ModelSpec
from typing import Dict, Any, Literal, Optional
from ..logger import logger
from ..config.config_manager import ConfigManager



class ModelManager:
    def __init__(self):
        self.llm_dict: Dict[str, Any] = {}               # dict to all llm models process object
        self.llm_dict_non_strict: Dict[str, Any] = {}
        self.tts_dict: Dict[str, Any] = {}
        self.tts_dict_non_strict: Dict[str, Any] = {}
        self.stt_dict: Dict[str, Any] = {}
        self.stt_dict_non_strict: Dict[str, Any] = {}
        self.embedding_dict: Dict[str, Any] = {}
        self.embedding_dict_non_strict: Dict[str, Any] = {}
        self.config_manager = ConfigManager()
        self.model_ready = False

    def get_model(self, model_name: str, _type: Literal["llm", "tts", "stt", "embedding"]) -> Optional[Any]:
        # Determine which dictionaries to use based on the type
        if _type == "llm":
            strict_dict = self.llm_dict
            non_strict_dict = self.llm_dict_non_strict
        elif _type == "tts":
            strict_dict = self.tts_dict
            non_strict_dict = self.tts_dict_non_strict
        elif _type == "stt":
            strict_dict = self.stt_dict
            non_strict_dict = self.stt_dict_non_strict
        elif _type == "embedding":
            strict_dict = self.embedding_dict
            non_strict_dict = self.embedding_dict_non_strict
        else:
            raise ValueError(f"Invalid model type: {_type}")

        # Check if the model exists in the strict dictionary
        if model_name in strict_dict:
            return strict_dict[model_name]

        # If not, check if there are any models in the non-strict dictionary
        if non_strict_dict:
            # Return the first model in the non-strict dictionary
            return next(iter(non_strict_dict.values()))

        # If no model is found, return None
        return None

    def _update_model(self, model_name: str, model_spec: ModelSpec, model_object: Any):
        if model_spec.model_type == "llm":
            self.llm_dict[model_name] = model_object
            if not model_spec.strict:
                self.llm_dict_non_strict[model_name] = model_object
        elif model_spec.model_type == "tts":
            self.tts_dict[model_name] = model_object
            if not model_spec.strict:
                self.tts_dict_non_strict[model_name] = model_object
        elif model_spec.model_type == "stt":
            self.stt_dict[model_name] = model_object
            if not model_spec.strict:
                self.stt_dict_non_strict[model_name] = model_object
        elif model_spec.model_type == "embedding":
            self.embedding_dict[model_name] = model_object
            if not model_spec.strict:
                self.embedding_dict_non_strict[model_name] = model_object


    def load_model(self, model_spec: ModelSpec):
        """
        model_spec is model specification coming from cli
        it might not have all the properties required for the model to be loaded
        the config_manager below contain all the models properties
        """
        # global config_manager, llm_dict, stt_dict, tts_dict

        # get the config from the yml
        model_name = model_spec.model_name
        model_config = self.config_manager.get_model_config(model_name)
        if not model_config:
            raise Exception(f"Model config for '{model_name}' not exist in ~/gallama/model_config.yaml")

        model_config.update({"model_name": model_name})

        # handle draft model
        if model_spec.draft_model_name and not model_spec.draft_model_id:
            draft_model_config = self.config_manager.get_model_config(model_spec.draft_model_name)
            model_config.update({
                "draft_model_id": model_spec.draft_model_id or draft_model_config["model_id"],
                "draft_model_name": model_spec.draft_model_name or draft_model_config["model_name"],
                "draft_gpus": model_spec.draft_gpus or draft_model_config["gpus"],
                "draft_cache_quant": model_spec.draft_cache_quant or draft_model_config["cache_quant"],
            })

        _default_model_spec = ModelSpec.from_dict(model_config)

        # Merge configurations that user pass in with default setting of the model
        model_spec = ModelSpec.from_merged_config(model_spec, _default_model_spec.model_dump())


        # load the model with config from the model_spec and yml. model_spec comes from cli
        if model_spec.backend in ["exllama", "llama_cpp", "transformers", "mlx_vlm"]:  # llm loading
            if model_spec.backend == "exllama":
                from gallama.backend.llm import ModelExllama as ModelClass
            elif model_spec.backend == "llama_cpp":
                from gallama.backend.llm import ModelLlamaCpp as ModelClass
            elif model_spec.backend == "transformers":
                from gallama.backend.llm import ModelTransformers as ModelClass
            elif model_spec.backend == "mlx_vlm":
                logger.info("frog mlx_vlm")
                from gallama.backend.llm import ModelMLXVLM as ModelClass
            else:
                raise Exception(f"Unknown backend: {model_spec.backend}")

            if model_spec.draft_model_id:
                draft_model_config = self.config_manager.get_model_config(model_spec.draft_model_name)
                if not draft_model_config:
                    raise Exception(
                        f"Model config for '{model_spec.draft_model_name}' not exist in ~/gallama/model_config.yaml")
            else:
                draft_model_config = {}

            logger.info(f"model_spec: {model_spec}")
            llm = ModelClass(model_spec=model_spec)

            # update dict
            self._update_model(
                model_name=model_name,
                model_spec=model_spec,
                model_object=llm
            )
        elif model_spec.backend == "embedding":  # embedding model
            from gallama.backend.embedding.embedding import EmbeddingModel

            self._update_model(
                model_name=model_name,
                model_spec=model_spec,
                model_object=EmbeddingModel(model_spec=model_spec)
            )


        elif model_spec.backend == "faster_whisper":  # embedding model
            from gallama.backend.stt import ASRProcessor, ASRFasterWhisper

            stt_base = ASRFasterWhisper(model_spec=model_spec)

            stt = ASRProcessor(asr=stt_base)

            # update dict
            self._update_model(
                model_name=model_name,
                model_spec=model_spec,
                model_object=stt
            )

        elif model_spec.backend == "mlx_whisper":  # embedding model
            from gallama.backend.stt import ASRProcessor, ASRMLXWhisper

            stt_base = ASRMLXWhisper(model_spec=model_spec)

            stt = ASRProcessor(asr=stt_base)

            # update dict
            self._update_model(
                model_name=model_name,
                model_spec=model_spec,
                model_object=stt
            )

        elif model_spec.backend == "gpt_sovits":  # embedding model
            from gallama.backend.tts import TTS_GPT_SoVITS
            tts = TTS_GPT_SoVITS(model_spec=model_spec)

            # update dict
            self._update_model(
                model_name=model_name,
                model_spec=model_spec,
                model_object=tts
            )
        elif model_spec.backend == "kokoro":  # embedding model
            from gallama.backend.tts import TTSKokoro
            tts = TTSKokoro(model_spec=model_spec)

            # update dict
            self._update_model(
                model_name=model_name,
                model_spec=model_spec,
                model_object=tts
            )

        else:
            raise Exception(f"Unknown backend: {model_spec.backend}")



        logger.info("Loaded: " + model_name)


from gallama.server_engine import ServerManager
import time
from gallama.logger.logger import get_logger
import logging
import zmq
import json
import threading
import traceback


server_manager = ServerManager()

def get_server_manager():
    if server_manager is None:
        raise RuntimeError("ServerManager not initialized")
    return server_manager



# set up logging
def start_log_receiver(zmq_url):
    context = zmq.Context()
    socket = context.socket(zmq.PULL)
    socket.bind("tcp://*:5555")

    # Initialize the logger for the receiver
    receiver_logger = get_logger(name="log_receiver", to_console=True, to_file=False, to_zmq=False)

    receiver_logger.info(f"Log receiver started on {zmq_url}")

    def receive_logs():
        while True:
            try:
                message = socket.recv_json()
                log_level = getattr(logging, message['level'].upper(), logging.INFO)

                receiver_logger.log(
                    level=log_level,
                    msg=f"{message['model']}:{message['port']} | {message['log']}"
                )
            except zmq.Again:
                # No message available, sleep for a short time
                time.sleep(0.1)
            except zmq.ZMQError as e:
                receiver_logger.error(f"ZMQ Error in log receiver: {e}")
            except json.JSONDecodeError as e:
                receiver_logger.error(f"JSON Decode Error in log receiver: {e}")
            except Exception as e:
                receiver_logger.error(f"Unexpected error in log receiver: {e}")
                receiver_logger.error(traceback.format_exc())

    # Start the receiver in a separate thread
    receiver_thread = threading.Thread(target=receive_logs, daemon=True)
    receiver_thread.start()

    return receiver_thread


# Start the log receiver in a separate thread
DEFAULT_ZMQ_URL = "tcp://127.0.0.1:5555"  # Using 5559 as a standard port for logging


# Initialize the logger for the manager
# Set to_console=True and to_zmq=False to avoid duplication
server_logger = get_logger(name="manager", to_console=True, to_zmq=False)

def get_server_logger():
    if server_logger is None:
        raise RuntimeError("ServerManager not initialized")
    return server_logger



# modified from https://github.com/yangdongchao/SoundStorm/blob/master/soundstorm/s1/AR/text_processing/phonemizer.py
# reference: https://github.com/lifeiteng/vall-e
import itertools
import re
from typing import Dict
from typing import List

import regex
from gruut import sentences
from gruut.const import Sentence
from gruut.const import Word
from AR.text_processing.symbols import SYMBOL_TO_ID


class GruutPhonemizer:
    def __init__(self, language: str):
        self._phonemizer = sentences
        self.lang = language
        self.symbol_to_id = SYMBOL_TO_ID
        self._special_cases_dict: Dict[str] = {
            r"\.\.\.": "... ",
            ";": "; ",
            ":": ": ",
            ",": ", ",
            r"\.": ". ",
            "!": "! ",
            r"\?": "? ",
            "—": "—",
            "…": "… ",
            "«": "«",
            "»": "»",
        }
        self._punctuation_regexp: str = (
            rf"([{''.join(self._special_cases_dict.keys())}])"
        )

    def _normalize_punctuation(self, text: str) -> str:
        text = regex.sub(rf"\pZ+{self._punctuation_regexp}", r"\1", text)
        text = regex.sub(rf"{self._punctuation_regexp}(\pL)", r"\1 \2", text)
        text = regex.sub(r"\pZ+", r" ", text)
        return text.strip()

    def _convert_punctuation(self, word: Word) -> str:
        if not word.phonemes:
            return ""
        if word.phonemes[0] in ["‖", "|"]:
            return word.text.strip()

        phonemes = "".join(word.phonemes)
        # remove modifier characters ˈˌː with regex
        phonemes = re.sub(r"[ˈˌː͡]", "", phonemes)
        return phonemes.strip()

    def phonemize(self, text: str, espeak: bool = False) -> str:
        text_to_phonemize: str = self._normalize_punctuation(text)
        sents: List[Sentence] = [
            sent
            for sent in self._phonemizer(text_to_phonemize, lang="en-us", espeak=espeak)
        ]
        words: List[str] = [
            self._convert_punctuation(word) for word in itertools.chain(*sents)
        ]
        return " ".join(words)

    def transform(self, phonemes):
        # convert phonemes to ids
        # dictionary is in symbols.py
        return [self.symbol_to_id[p] for p in phonemes if p in self.symbol_to_id.keys()]


if __name__ == "__main__":
    phonemizer = GruutPhonemizer("en-us")
    # text -> IPA
    phonemes = phonemizer.phonemize("Hello, wor-ld ?")
    print("phonemes:", phonemes)
    print("len(phonemes):", len(phonemes))
    phoneme_ids = phonemizer.transform(phonemes)
    print("phoneme_ids:", phoneme_ids)
    print("len(phoneme_ids):", len(phoneme_ids))



from .model_management import download_model_from_hf, log_model_status
from .mixture_of_agents import handle_mixture_of_agent_request
from .request_handler import forward_request, create_options_response
from .server_manager import ServerManager


from copy import deepcopy
import math
import os, sys, gc
import random
import traceback
from tqdm import tqdm
now_dir = os.getcwd()
sys.path.append(now_dir)
current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(current_dir)
gpt_sovits_dir = os.path.dirname(parent_dir)
# Add all relevant directories to Python path
for directory in [current_dir, parent_dir, gpt_sovits_dir]:
    if directory not in sys.path:
        sys.path.append(directory)
import ffmpeg
import os
from typing import Generator, List, Tuple, Union
import numpy as np
import torch
import torch.nn.functional as F
import yaml
from transformers import AutoModelForMaskedLM, AutoTokenizer

from AR.models.t2s_lightning_module import Text2SemanticLightningModule
from feature_extractor.cnhubert import CNHubert
from module.models import SynthesizerTrn
import librosa
from time import time as ttime
from tools.i18n.i18n import I18nAuto, scan_language_list
from tools.my_utils import load_audio
from module.mel_processing import spectrogram_torch
from TTS_infer_pack.text_segmentation_method import splits
from TTS_infer_pack.TextPreprocessor import TextPreprocessor
language=os.environ.get("language","Auto")
language=sys.argv[-1] if sys.argv[-1] in scan_language_list() else language
i18n = I18nAuto(language=language)

# configs/tts_infer.yaml
"""
custom:
  bert_base_path: GPT_SoVITS/pretrained_models/chinese-roberta-wwm-ext-large
  cnhuhbert_base_path: GPT_SoVITS/pretrained_models/chinese-hubert-base
  device: cpu
  is_half: false
  t2s_weights_path: GPT_SoVITS/pretrained_models/gsv-v2final-pretrained/s1bert25hz-5kh-longer-epoch=12-step=369668.ckpt
  vits_weights_path: GPT_SoVITS/pretrained_models/gsv-v2final-pretrained/s2G2333k.pth
  version: v2
default:
  bert_base_path: GPT_SoVITS/pretrained_models/chinese-roberta-wwm-ext-large
  cnhuhbert_base_path: GPT_SoVITS/pretrained_models/chinese-hubert-base
  device: cpu
  is_half: false
  t2s_weights_path: GPT_SoVITS/pretrained_models/s1bert25hz-2kh-longer-epoch=68e-step=50232.ckpt
  vits_weights_path: GPT_SoVITS/pretrained_models/s2G488k.pth
  version: v1
default_v2:
  bert_base_path: GPT_SoVITS/pretrained_models/chinese-roberta-wwm-ext-large
  cnhuhbert_base_path: GPT_SoVITS/pretrained_models/chinese-hubert-base
  device: cpu
  is_half: false
  t2s_weights_path: GPT_SoVITS/pretrained_models/gsv-v2final-pretrained/s1bert25hz-5kh-longer-epoch=12-step=369668.ckpt
  vits_weights_path: GPT_SoVITS/pretrained_models/gsv-v2final-pretrained/s2G2333k.pth
  version: v2
"""

def set_seed(seed:int):
    seed = int(seed)
    seed = seed if seed != -1 else random.randrange(1 << 32)
    print(f"Set seed to {seed}")
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    try:
        if torch.cuda.is_available():
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            # torch.backends.cudnn.deterministic = True
            # torch.backends.cudnn.benchmark = False
            # torch.backends.cudnn.enabled = True
            # 开启后会影响精度
            torch.backends.cuda.matmul.allow_tf32 = False
            torch.backends.cudnn.allow_tf32 = False
    except:
        pass
    return seed
    
class TTS_Config:
    default_configs={
        "default":{
                "device": "cpu",
                "is_half": False,
                "version": "v1",
                "t2s_weights_path": "GPT_SoVITS/pretrained_models/s1bert25hz-2kh-longer-epoch=68e-step=50232.ckpt",
                "vits_weights_path": "GPT_SoVITS/pretrained_models/s2G488k.pth",
                "cnhuhbert_base_path": "GPT_SoVITS/pretrained_models/chinese-hubert-base",
                "bert_base_path": "GPT_SoVITS/pretrained_models/chinese-roberta-wwm-ext-large",
            },
        "default_v2":{
                "device": "cpu",
                "is_half": False,
                "version": "v2",
                "t2s_weights_path": "GPT_SoVITS/pretrained_models/gsv-v2final-pretrained/s1bert25hz-5kh-longer-epoch=12-step=369668.ckpt",
                "vits_weights_path": "GPT_SoVITS/pretrained_models/gsv-v2final-pretrained/s2G2333k.pth",
                "cnhuhbert_base_path": "GPT_SoVITS/pretrained_models/chinese-hubert-base",
                "bert_base_path": "GPT_SoVITS/pretrained_models/chinese-roberta-wwm-ext-large",
            },
    }
    configs:dict = None
    v1_languages:list = ["auto", "en", "zh", "ja",  "all_zh", "all_ja"]
    v2_languages:list = ["auto", "auto_yue", "en", "zh", "ja", "yue", "ko", "all_zh", "all_ja", "all_yue", "all_ko"]
    languages:list = v2_languages
    # "all_zh",#全部按中文识别
    # "en",#全部按英文识别#######不变
    # "all_ja",#全部按日文识别
    # "all_yue",#全部按中文识别
    # "all_ko",#全部按韩文识别
    # "zh",#按中英混合识别####不变
    # "ja",#按日英混合识别####不变
    # "yue",#按粤英混合识别####不变
    # "ko",#按韩英混合识别####不变
    # "auto",#多语种启动切分识别语种
    # "auto_yue",#多语种启动切分识别语种

    def __init__(self, configs: Union[dict, str]=None):
        
        # 设置默认配置文件路径
        configs_base_path:str = "GPT_SoVITS/configs/"
        os.makedirs(configs_base_path, exist_ok=True)
        self.configs_path:str = os.path.join(configs_base_path, "tts_infer.yaml")
        
        if configs in ["", None]:
            if not os.path.exists(self.configs_path):
                self.save_configs()
                print(f"Create default config file at {self.configs_path}")
            configs:dict = deepcopy(self.default_configs)
        
        if isinstance(configs, str):
            self.configs_path = configs
            configs:dict = self._load_configs(self.configs_path)

        assert isinstance(configs, dict)
        version = configs.get("version", "v2").lower()
        assert version in ["v1", "v2"]
        self.default_configs["default"] = configs.get("default", self.default_configs["default"])
        self.default_configs["default_v2"] = configs.get("default_v2", self.default_configs["default_v2"])

        default_config_key = "default"if version=="v1" else "default_v2"
        self.configs:dict = configs.get("custom", deepcopy(self.default_configs[default_config_key]))
        
        
        self.device = self.configs.get("device", torch.device("cpu"))
        self.is_half = self.configs.get("is_half", False)
        self.version = version
        self.t2s_weights_path = self.configs.get("t2s_weights_path", None)
        self.vits_weights_path = self.configs.get("vits_weights_path", None)
        self.bert_base_path = self.configs.get("bert_base_path", None)
        self.cnhuhbert_base_path = self.configs.get("cnhuhbert_base_path", None)
        self.languages = self.v2_languages if self.version=="v2" else self.v1_languages

        
        if (self.t2s_weights_path in [None, ""]) or (not os.path.exists(self.t2s_weights_path)):
            self.t2s_weights_path = self.default_configs[default_config_key]['t2s_weights_path']
            print(f"fall back to default t2s_weights_path: {self.t2s_weights_path}")
        if (self.vits_weights_path in [None, ""]) or (not os.path.exists(self.vits_weights_path)):
            self.vits_weights_path = self.default_configs[default_config_key]['vits_weights_path']
            print(f"fall back to default vits_weights_path: {self.vits_weights_path}")
        if (self.bert_base_path in [None, ""]) or (not os.path.exists(self.bert_base_path)):
            self.bert_base_path = self.default_configs[default_config_key]['bert_base_path']
            print(f"fall back to default bert_base_path: {self.bert_base_path}")
        if (self.cnhuhbert_base_path in [None, ""]) or (not os.path.exists(self.cnhuhbert_base_path)):
            self.cnhuhbert_base_path = self.default_configs[default_config_key]['cnhuhbert_base_path']
            print(f"fall back to default cnhuhbert_base_path: {self.cnhuhbert_base_path}")
        self.update_configs()
        
        
        self.max_sec = None
        self.hz:int = 50
        self.semantic_frame_rate:str = "25hz"
        self.segment_size:int = 20480
        self.filter_length:int = 2048
        self.sampling_rate:int = 32000
        self.hop_length:int = 640
        self.win_length:int = 2048
        self.n_speakers:int = 300


            
    def _load_configs(self, configs_path: str)->dict:
        if os.path.exists(configs_path):
            ...
        else:
            print(i18n("路径不存在,使用默认配置"))
            self.save_configs(configs_path)
        with open(configs_path, 'r') as f:
            configs = yaml.load(f, Loader=yaml.FullLoader)
    
        return configs

    def save_configs(self, configs_path:str=None)->None:
        configs=deepcopy(self.default_configs)
        if self.configs is not None:
            configs["custom"] = self.update_configs()

        if configs_path is None:
            configs_path = self.configs_path
        with open(configs_path, 'w') as f:
            yaml.dump(configs, f)

    def update_configs(self):
        self.config = {
            "device"             : str(self.device),
            "is_half"            : self.is_half,
            "version"            : self.version,
            "t2s_weights_path"   : self.t2s_weights_path,
            "vits_weights_path"  : self.vits_weights_path,
            "bert_base_path"     : self.bert_base_path,
            "cnhuhbert_base_path": self.cnhuhbert_base_path,
        }
        return self.config

    def update_version(self, version:str)->None:
        self.version = version
        self.languages = self.v2_languages if self.version=="v2" else self.v1_languages
            
    def __str__(self):
        self.configs = self.update_configs()
        string = "TTS Config".center(100, '-') + '\n'
        for k, v in self.configs.items():
            string += f"{str(k).ljust(20)}: {str(v)}\n"
        string += "-" * 100 + '\n'
        return string
    
    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.configs_path)

    def __eq__(self, other):
        return isinstance(other, TTS_Config) and self.configs_path == other.configs_path


class TTS:
    def __init__(self, configs: Union[dict, str, TTS_Config]):
        if isinstance(configs, TTS_Config):
            self.configs = configs
        else:
            self.configs:TTS_Config = TTS_Config(configs)
        
        self.t2s_model:Text2SemanticLightningModule = None
        self.vits_model:SynthesizerTrn = None
        self.bert_tokenizer:AutoTokenizer = None
        self.bert_model:AutoModelForMaskedLM = None
        self.cnhuhbert_model:CNHubert = None
        
        self._init_models()
        
        self.text_preprocessor:TextPreprocessor = \
                            TextPreprocessor(self.bert_model, 
                                            self.bert_tokenizer, 
                                            self.configs.device)
        
        
        self.prompt_cache:dict = {
            "ref_audio_path" : None,
            "prompt_semantic": None,
            "refer_spec"     : [],
            "prompt_text"    : None,
            "prompt_lang"    : None,
            "phones"         : None,
            "bert_features"  : None,
            "norm_text"      : None,
            "aux_ref_audio_paths": [],
        }
        
        
        self.stop_flag:bool = False
        self.precision:torch.dtype = torch.float16 if self.configs.is_half else torch.float32

    def _init_models(self,):
        self.init_t2s_weights(self.configs.t2s_weights_path)
        self.init_vits_weights(self.configs.vits_weights_path)
        self.init_bert_weights(self.configs.bert_base_path)
        self.init_cnhuhbert_weights(self.configs.cnhuhbert_base_path)
        # self.enable_half_precision(self.configs.is_half)
        
        
        
    def init_cnhuhbert_weights(self, base_path: str):
        print(f"Loading CNHuBERT weights from {base_path}")
        self.cnhuhbert_model = CNHubert(base_path)
        self.cnhuhbert_model=self.cnhuhbert_model.eval()
        self.cnhuhbert_model = self.cnhuhbert_model.to(self.configs.device)
        if self.configs.is_half and str(self.configs.device)!="cpu":
            self.cnhuhbert_model = self.cnhuhbert_model.half()
        
        
        
    def init_bert_weights(self, base_path: str):
        print(f"Loading BERT weights from {base_path}")
        self.bert_tokenizer = AutoTokenizer.from_pretrained(base_path)
        self.bert_model = AutoModelForMaskedLM.from_pretrained(base_path)
        self.bert_model=self.bert_model.eval()
        self.bert_model = self.bert_model.to(self.configs.device)
        if self.configs.is_half and str(self.configs.device)!="cpu":
            self.bert_model = self.bert_model.half()
        
    def init_vits_weights(self, weights_path: str):
        print(f"Loading VITS weights from {weights_path}")
        self.configs.vits_weights_path = weights_path
        dict_s2 = torch.load(weights_path, map_location=self.configs.device)
        hps = dict_s2["config"]
        if dict_s2['weight']['enc_p.text_embedding.weight'].shape[0] == 322:
            self.configs.update_version("v1")
        else:
            self.configs.update_version("v2")
        self.configs.save_configs()
        
        hps["model"]["version"] = self.configs.version
        self.configs.filter_length = hps["data"]["filter_length"]
        self.configs.segment_size = hps["train"]["segment_size"]
        self.configs.sampling_rate = hps["data"]["sampling_rate"]       
        self.configs.hop_length = hps["data"]["hop_length"]
        self.configs.win_length = hps["data"]["win_length"]
        self.configs.n_speakers = hps["data"]["n_speakers"]
        self.configs.semantic_frame_rate = "25hz"
        kwargs = hps["model"]
        vits_model = SynthesizerTrn(
            self.configs.filter_length // 2 + 1,
            self.configs.segment_size // self.configs.hop_length,
            n_speakers=self.configs.n_speakers,
            **kwargs
        )

        if hasattr(vits_model, "enc_q"):
            del vits_model.enc_q
            
        vits_model = vits_model.to(self.configs.device)
        vits_model = vits_model.eval()
        vits_model.load_state_dict(dict_s2["weight"], strict=False)
        self.vits_model = vits_model
        if self.configs.is_half and str(self.configs.device)!="cpu":
            self.vits_model = self.vits_model.half()

        
    def init_t2s_weights(self, weights_path: str):
        print(f"Loading Text2Semantic weights from {weights_path}")
        self.configs.t2s_weights_path = weights_path
        self.configs.save_configs()
        self.configs.hz = 50
        dict_s1 = torch.load(weights_path, map_location=self.configs.device)
        config = dict_s1["config"]
        self.configs.max_sec = config["data"]["max_sec"]
        t2s_model = Text2SemanticLightningModule(config, "****", is_train=False)
        t2s_model.load_state_dict(dict_s1["weight"])
        t2s_model = t2s_model.to(self.configs.device)
        t2s_model = t2s_model.eval()
        self.t2s_model = t2s_model
        if self.configs.is_half and str(self.configs.device)!="cpu":
            self.t2s_model = self.t2s_model.half()
        
    def enable_half_precision(self, enable: bool = True, save: bool = True):
        '''
            To enable half precision for the TTS model.
            Args:
                enable: bool, whether to enable half precision.
                
        '''
        if str(self.configs.device) == "cpu" and enable:
            print("Half precision is not supported on CPU.")
            return
        
        self.configs.is_half = enable
        self.precision = torch.float16 if enable else torch.float32
        if save:
            self.configs.save_configs()
        if enable:
            if self.t2s_model is not None:
                self.t2s_model =self.t2s_model.half()
            if self.vits_model is not None:
                self.vits_model = self.vits_model.half()
            if self.bert_model is not None:
                self.bert_model =self.bert_model.half()
            if self.cnhuhbert_model is not None:
                self.cnhuhbert_model = self.cnhuhbert_model.half()
        else:
            if self.t2s_model is not None:
                self.t2s_model = self.t2s_model.float()
            if self.vits_model is not None:
                self.vits_model = self.vits_model.float()
            if self.bert_model is not None:
                self.bert_model = self.bert_model.float()
            if self.cnhuhbert_model is not None:
                self.cnhuhbert_model = self.cnhuhbert_model.float()
                
    def set_device(self, device: torch.device, save: bool = True):
        '''
            To set the device for all models.
            Args:
                device: torch.device, the device to use for all models.
        '''
        self.configs.device = device
        if save:
            self.configs.save_configs()
        if self.t2s_model is not None:
            self.t2s_model = self.t2s_model.to(device)
        if self.vits_model is not None:
            self.vits_model = self.vits_model.to(device)
        if self.bert_model is not None:
            self.bert_model = self.bert_model.to(device)
        if self.cnhuhbert_model is not None:
            self.cnhuhbert_model = self.cnhuhbert_model.to(device)
        
    def set_ref_audio(self, ref_audio_path:str):
        '''
            To set the reference audio for the TTS model, 
                including the prompt_semantic and refer_spepc.
            Args:
                ref_audio_path: str, the path of the reference audio.
        '''
        self._set_prompt_semantic(ref_audio_path)
        self._set_ref_spec(ref_audio_path)
        self._set_ref_audio_path(ref_audio_path)
        
    def _set_ref_audio_path(self, ref_audio_path):
        self.prompt_cache["ref_audio_path"] = ref_audio_path 

    def _set_ref_spec(self, ref_audio_path):
        spec = self._get_ref_spec(ref_audio_path)
        if self.prompt_cache["refer_spec"] in [[],None]:
            self.prompt_cache["refer_spec"]=[spec]
        else:
            self.prompt_cache["refer_spec"][0] = spec

    def _get_ref_spec(self, ref_audio_path):
        audio = load_audio(ref_audio_path, int(self.configs.sampling_rate))
        audio = torch.FloatTensor(audio)
        maxx=audio.abs().max()
        if(maxx>1):audio/=min(2,maxx)
        audio_norm = audio
        audio_norm = audio_norm.unsqueeze(0)
        spec = spectrogram_torch(
            audio_norm,
            self.configs.filter_length,
            self.configs.sampling_rate,
            self.configs.hop_length,
            self.configs.win_length,
            center=False,
        )
        spec = spec.to(self.configs.device)
        if self.configs.is_half:
            spec = spec.half()
        return spec

    def _set_prompt_semantic(self, ref_wav_path:str):
        zero_wav = np.zeros(
            int(self.configs.sampling_rate * 0.3),
            dtype=np.float16 if self.configs.is_half else np.float32,
        )
        with torch.no_grad():
            wav16k, sr = librosa.load(ref_wav_path, sr=16000)
            if (wav16k.shape[0] > 160000 or wav16k.shape[0] < 48000):
                raise OSError(i18n("参考音频在3~10秒范围外，请更换！"))
            wav16k = torch.from_numpy(wav16k)
            zero_wav_torch = torch.from_numpy(zero_wav)
            wav16k = wav16k.to(self.configs.device)
            zero_wav_torch = zero_wav_torch.to(self.configs.device)
            if self.configs.is_half:
                wav16k = wav16k.half()
                zero_wav_torch = zero_wav_torch.half()

            wav16k = torch.cat([wav16k, zero_wav_torch])
            hubert_feature = self.cnhuhbert_model.model(wav16k.unsqueeze(0))[
                "last_hidden_state"
            ].transpose(
                1, 2
            )  # .float()
            codes = self.vits_model.extract_latent(hubert_feature)
    
            prompt_semantic = codes[0, 0].to(self.configs.device)
            self.prompt_cache["prompt_semantic"] = prompt_semantic
    
    def batch_sequences(self, sequences: List[torch.Tensor], axis: int = 0, pad_value: int = 0, max_length:int=None):
        seq = sequences[0]
        ndim = seq.dim()
        if axis < 0:
            axis += ndim
        dtype:torch.dtype = seq.dtype
        pad_value = torch.tensor(pad_value, dtype=dtype)
        seq_lengths = [seq.shape[axis] for seq in sequences]
        if max_length is None:
            max_length = max(seq_lengths)
        else:
            max_length = max(seq_lengths) if max_length < max(seq_lengths) else max_length

        padded_sequences = []
        for seq, length in zip(sequences, seq_lengths):
            padding = [0] * axis + [0, max_length - length] + [0] * (ndim - axis - 1)
            padded_seq = torch.nn.functional.pad(seq, padding, value=pad_value)
            padded_sequences.append(padded_seq)
        batch = torch.stack(padded_sequences)
        return batch
    
    def to_batch(self, data:list, 
                 prompt_data:dict=None, 
                 batch_size:int=5, 
                 threshold:float=0.75, 
                 split_bucket:bool=True, 
                 device:torch.device=torch.device("cpu"),
                 precision:torch.dtype=torch.float32,
                 ):
        _data:list = []
        index_and_len_list = []
        for idx, item in enumerate(data):
            norm_text_len = len(item["norm_text"])
            index_and_len_list.append([idx, norm_text_len])

        batch_index_list = []
        if split_bucket:
            index_and_len_list.sort(key=lambda x: x[1])
            index_and_len_list = np.array(index_and_len_list, dtype=np.int64)            
            
            batch_index_list_len = 0
            pos = 0
            while pos <index_and_len_list.shape[0]:
                # batch_index_list.append(index_and_len_list[pos:min(pos+batch_size,len(index_and_len_list))])
                pos_end = min(pos+batch_size,index_and_len_list.shape[0])
                while pos < pos_end:
                    batch=index_and_len_list[pos:pos_end, 1].astype(np.float32)
                    score=batch[(pos_end-pos)//2]/(batch.mean()+1e-8)
                    if (score>=threshold) or (pos_end-pos==1):
                        batch_index=index_and_len_list[pos:pos_end, 0].tolist()
                        batch_index_list_len += len(batch_index)
                        batch_index_list.append(batch_index)
                        pos = pos_end
                        break
                    pos_end=pos_end-1
            
            assert batch_index_list_len == len(data)
            
        else:
            for i in range(len(data)):
                if i%batch_size == 0:
                    batch_index_list.append([])
                batch_index_list[-1].append(i)

                
        for batch_idx, index_list in enumerate(batch_index_list):
            item_list = [data[idx] for idx in index_list]
            phones_list = []
            phones_len_list = []
            # bert_features_list = []
            all_phones_list = []
            all_phones_len_list = []
            all_bert_features_list = []
            norm_text_batch = []
            all_bert_max_len = 0
            all_phones_max_len = 0
            for item in item_list:
                if prompt_data is not None:
                    all_bert_features = torch.cat([prompt_data["bert_features"], item["bert_features"]], 1)\
                                                .to(dtype=precision, device=device)
                    all_phones = torch.LongTensor(prompt_data["phones"]+item["phones"]).to(device)
                    phones = torch.LongTensor(item["phones"]).to(device)
                    # norm_text = prompt_data["norm_text"]+item["norm_text"]
                else:
                    all_bert_features = item["bert_features"]\
                                            .to(dtype=precision, device=device)
                    phones = torch.LongTensor(item["phones"]).to(device)
                    all_phones = phones
                    # norm_text = item["norm_text"]

                all_bert_max_len = max(all_bert_max_len, all_bert_features.shape[-1])
                all_phones_max_len = max(all_phones_max_len, all_phones.shape[-1])
                
                phones_list.append(phones)
                phones_len_list.append(phones.shape[-1])
                all_phones_list.append(all_phones)
                all_phones_len_list.append(all_phones.shape[-1])
                all_bert_features_list.append(all_bert_features)
                norm_text_batch.append(item["norm_text"])
                
            phones_batch = phones_list
            all_phones_batch = all_phones_list
            all_bert_features_batch = all_bert_features_list
            
            
            max_len = max(all_bert_max_len, all_phones_max_len)
            # phones_batch = self.batch_sequences(phones_list, axis=0, pad_value=0, max_length=max_len)
            #### 直接对phones和bert_features进行pad。（padding策略会影响T2S模型生成的结果，但不直接影响复读概率。影响复读概率的主要因素是mask的策略）
            # all_phones_batch = self.batch_sequences(all_phones_list, axis=0, pad_value=0, max_length=max_len)
            # all_bert_features_batch = all_bert_features_list
            # all_bert_features_batch = torch.zeros((len(all_bert_features_list), 1024, max_len), dtype=precision, device=device)
            # for idx, item in enumerate(all_bert_features_list):
            #     all_bert_features_batch[idx, :, : item.shape[-1]] = item
            
            # #### 先对phones进行embedding、对bert_features进行project，再pad到相同长度，（padding策略会影响T2S模型生成的结果，但不直接影响复读概率。影响复读概率的主要因素是mask的策略）
            # all_phones_list = [self.t2s_model.model.ar_text_embedding(item.to(self.t2s_model.device)) for item in all_phones_list]
            # all_phones_list = [F.pad(item,(0,0,0,max_len-item.shape[0]),value=0) for item in all_phones_list]
            # all_phones_batch = torch.stack(all_phones_list, dim=0)
            
            # all_bert_features_list = [self.t2s_model.model.bert_proj(item.to(self.t2s_model.device).transpose(0, 1)) for item in all_bert_features_list]
            # all_bert_features_list = [F.pad(item,(0,0,0,max_len-item.shape[0]), value=0) for item in all_bert_features_list]
            # all_bert_features_batch = torch.stack(all_bert_features_list, dim=0)
            
            batch = {
                "phones": phones_batch,
                "phones_len": torch.LongTensor(phones_len_list).to(device),
                "all_phones": all_phones_batch,
                "all_phones_len": torch.LongTensor(all_phones_len_list).to(device),
                "all_bert_features": all_bert_features_batch,
                "norm_text": norm_text_batch,
                "max_len": max_len,
            }
            _data.append(batch)
        
        return _data, batch_index_list
    
    def recovery_order(self, data:list, batch_index_list:list)->list:
        '''
        Recovery the order of the audio according to the batch_index_list.
        
        Args:
            data (List[list(np.ndarray)]): the out of order audio .
            batch_index_list (List[list[int]]): the batch index list.
        
        Returns:
            list (List[np.ndarray]): the data in the original order.
        '''
        length = len(sum(batch_index_list, []))
        _data = [None]*length
        for i, index_list in enumerate(batch_index_list):
            for j, index in enumerate(index_list):
                _data[index] = data[i][j]
        return _data

    def stop(self,):
        '''
        Stop the inference process.
        '''
        self.stop_flag = True
    
    @torch.no_grad()
    def run(self, inputs:dict):
        """
        Text to speech inference.
        
        Args:
            inputs (dict): 
                {
                    "text": "",                   # str.(required) text to be synthesized
                    "text_lang: "",               # str.(required) language of the text to be synthesized
                    "ref_audio_path": "",         # str.(required) reference audio path
                    "aux_ref_audio_paths": [],    # list.(optional) auxiliary reference audio paths for multi-speaker tone fusion
                    "prompt_text": "",            # str.(optional) prompt text for the reference audio
                    "prompt_lang": "",            # str.(required) language of the prompt text for the reference audio
                    "top_k": 5,                   # int. top k sampling
                    "top_p": 1,                   # float. top p sampling
                    "temperature": 1,             # float. temperature for sampling
                    "text_split_method": "cut0",  # str. text split method, see text_segmentation_method.py for details.
                    "batch_size": 1,              # int. batch size for inference
                    "batch_threshold": 0.75,      # float. threshold for batch splitting.
                    "split_bucket: True,          # bool. whether to split the batch into multiple buckets.
                    "return_fragment": False,     # bool. step by step return the audio fragment.
                    "speed_factor":1.0,           # float. control the speed of the synthesized audio.
                    "fragment_interval":0.3,      # float. to control the interval of the audio fragment.
                    "seed": -1,                   # int. random seed for reproducibility.
                    "parallel_infer": True,       # bool. whether to use parallel inference.
                    "repetition_penalty": 1.35    # float. repetition penalty for T2S model.
                }
        returns:
            Tuple[int, np.ndarray]: sampling rate and audio data.
        """
        ########## variables initialization ###########
        self.stop_flag:bool = False
        text:str = inputs.get("text", "")
        text_lang:str = inputs.get("text_lang", "")
        ref_audio_path:str = inputs.get("ref_audio_path", "")
        aux_ref_audio_paths:list = inputs.get("aux_ref_audio_paths", [])
        prompt_text:str = inputs.get("prompt_text", "")
        prompt_lang:str = inputs.get("prompt_lang", "")
        top_k:int = inputs.get("top_k", 5)
        top_p:float = inputs.get("top_p", 1)
        temperature:float = inputs.get("temperature", 1)
        text_split_method:str = inputs.get("text_split_method", "cut0")
        batch_size = inputs.get("batch_size", 1)
        batch_threshold = inputs.get("batch_threshold", 0.75)
        speed_factor = inputs.get("speed_factor", 1.0)
        split_bucket = inputs.get("split_bucket", True)
        return_fragment = inputs.get("return_fragment", False)
        fragment_interval = inputs.get("fragment_interval", 0.3)
        seed = inputs.get("seed", -1)
        seed = -1 if seed in ["", None] else seed
        actual_seed = set_seed(seed)
        parallel_infer = inputs.get("parallel_infer", True)
        repetition_penalty = inputs.get("repetition_penalty", 1.35)

        if parallel_infer:
            print(i18n("并行推理模式已开启"))
            self.t2s_model.model.infer_panel = self.t2s_model.model.infer_panel_batch_infer
        else:
            print(i18n("并行推理模式已关闭"))
            self.t2s_model.model.infer_panel = self.t2s_model.model.infer_panel_naive_batched

        if return_fragment:
            print(i18n("分段返回模式已开启"))
            if split_bucket:
                split_bucket = False
                print(i18n("分段返回模式不支持分桶处理，已自动关闭分桶处理"))

        if split_bucket and speed_factor==1.0:
            print(i18n("分桶处理模式已开启"))
        elif speed_factor!=1.0:
            print(i18n("语速调节不支持分桶处理，已自动关闭分桶处理"))
            split_bucket = False
        else:
            print(i18n("分桶处理模式已关闭"))

        if fragment_interval<0.01:
            fragment_interval = 0.01
            print(i18n("分段间隔过小，已自动设置为0.01"))

        no_prompt_text = False
        if prompt_text in [None, ""]:
            no_prompt_text = True

        assert text_lang in self.configs.languages
        if not no_prompt_text:
            assert prompt_lang in self.configs.languages

        if ref_audio_path in [None, ""] and \
            ((self.prompt_cache["prompt_semantic"] is None) or (self.prompt_cache["refer_spec"] in [None, []])):
            raise ValueError("ref_audio_path cannot be empty, when the reference audio is not set using set_ref_audio()")

        ###### setting reference audio and prompt text preprocessing ########
        t0 = ttime()
        if (ref_audio_path is not None) and (ref_audio_path != self.prompt_cache["ref_audio_path"]):
            if not os.path.exists(ref_audio_path):
                raise ValueError(f"{ref_audio_path} not exists")
            self.set_ref_audio(ref_audio_path)
            
        aux_ref_audio_paths = aux_ref_audio_paths if aux_ref_audio_paths is not None else []
        paths = set(aux_ref_audio_paths)&set(self.prompt_cache["aux_ref_audio_paths"])
        if not (len(list(paths)) == len(aux_ref_audio_paths) == len(self.prompt_cache["aux_ref_audio_paths"])):
            self.prompt_cache["aux_ref_audio_paths"] = aux_ref_audio_paths
            self.prompt_cache["refer_spec"] = [self.prompt_cache["refer_spec"][0]]
            for path in aux_ref_audio_paths:
                if path in [None, ""]:
                    continue
                if not os.path.exists(path):
                    print(i18n("音频文件不存在，跳过：{}").format(path))
                    continue
                self.prompt_cache["refer_spec"].append(self._get_ref_spec(path))
                
        if not no_prompt_text:
            prompt_text = prompt_text.strip("\n")
            if (prompt_text[-1] not in splits): prompt_text += "。" if prompt_lang != "en" else "."
            print(i18n("实际输入的参考文本:"), prompt_text)
            if self.prompt_cache["prompt_text"] != prompt_text:
                self.prompt_cache["prompt_text"] = prompt_text
                self.prompt_cache["prompt_lang"] = prompt_lang
                phones, bert_features, norm_text = \
                    self.text_preprocessor.segment_and_extract_feature_for_text(
                                                                        prompt_text, 
                                                                        prompt_lang,
                                                                        self.configs.version)
                self.prompt_cache["phones"] = phones
                self.prompt_cache["bert_features"] = bert_features
                self.prompt_cache["norm_text"] = norm_text




        ###### text preprocessing ########
        t1 = ttime()
        data:list = None
        if not return_fragment:
            data = self.text_preprocessor.preprocess(text, text_lang, text_split_method, self.configs.version)
            if len(data) == 0:
                yield self.configs.sampling_rate, np.zeros(int(self.configs.sampling_rate),
                                                            dtype=np.int16)
                return

            batch_index_list:list = None
            data, batch_index_list = self.to_batch(data, 
                                prompt_data=self.prompt_cache if not no_prompt_text else None, 
                                batch_size=batch_size, 
                                threshold=batch_threshold,
                                split_bucket=split_bucket,
                                device=self.configs.device,
                                precision=self.precision
                                )
        else:
            print(i18n("############ 切分文本 ############"))
            texts = self.text_preprocessor.pre_seg_text(text, text_lang, text_split_method)
            data = []
            for i in range(len(texts)):
                if i%batch_size == 0:
                    data.append([])
                data[-1].append(texts[i])
            
            def make_batch(batch_texts):
                batch_data = []
                print(i18n("############ 提取文本Bert特征 ############"))
                for text in tqdm(batch_texts):
                    phones, bert_features, norm_text = self.text_preprocessor.segment_and_extract_feature_for_text(text, text_lang, self.configs.version)
                    if phones is None:
                        continue
                    res={
                        "phones": phones,
                        "bert_features": bert_features,
                        "norm_text": norm_text,
                    }
                    batch_data.append(res)
                if len(batch_data) == 0:
                    return None
                batch, _ = self.to_batch(batch_data, 
                            prompt_data=self.prompt_cache if not no_prompt_text else None, 
                            batch_size=batch_size, 
                            threshold=batch_threshold,
                            split_bucket=False,
                            device=self.configs.device,
                            precision=self.precision
                            )
                return batch[0]


        t2 = ttime()
        try:
            print("############ 推理 ############")
            ###### inference ######
            t_34 = 0.0
            t_45 = 0.0
            audio = []
            for item in data:
                t3 = ttime()
                if return_fragment:
                    item = make_batch(item)
                    if item is None:
                        continue

                batch_phones:List[torch.LongTensor] = item["phones"]
                # batch_phones:torch.LongTensor = item["phones"]
                batch_phones_len:torch.LongTensor = item["phones_len"]
                all_phoneme_ids:torch.LongTensor = item["all_phones"]
                all_phoneme_lens:torch.LongTensor  = item["all_phones_len"]
                all_bert_features:torch.LongTensor = item["all_bert_features"]
                norm_text:str = item["norm_text"]
                max_len = item["max_len"]

                print(i18n("前端处理后的文本(每句):"), norm_text)
                if no_prompt_text :
                    prompt = None
                else:
                    prompt = self.prompt_cache["prompt_semantic"].expand(len(all_phoneme_ids), -1).to(self.configs.device)


                pred_semantic_list, idx_list = self.t2s_model.model.infer_panel(
                    all_phoneme_ids,
                    all_phoneme_lens,
                    prompt,
                    all_bert_features,
                    # prompt_phone_len=ph_offset,
                    top_k=top_k,
                    top_p=top_p,
                    temperature=temperature,
                    early_stop_num=self.configs.hz * self.configs.max_sec,
                    max_len=max_len,
                    repetition_penalty=repetition_penalty,
                )
                t4 = ttime()
                t_34 += t4 - t3

                refer_audio_spec:torch.Tensor = [item.to(dtype=self.precision, device=self.configs.device) for item in self.prompt_cache["refer_spec"]]
                                                    

                batch_audio_fragment = []
            
                # ## vits并行推理 method 1
                # pred_semantic_list = [item[-idx:] for item, idx in zip(pred_semantic_list, idx_list)]
                # pred_semantic_len = torch.LongTensor([item.shape[0] for item in pred_semantic_list]).to(self.configs.device)
                # pred_semantic = self.batch_sequences(pred_semantic_list, axis=0, pad_value=0).unsqueeze(0)
                # max_len = 0
                # for i in range(0, len(batch_phones)):
                #     max_len = max(max_len, batch_phones[i].shape[-1])
                # batch_phones = self.batch_sequences(batch_phones, axis=0, pad_value=0, max_length=max_len)
                # batch_phones = batch_phones.to(self.configs.device)
                # batch_audio_fragment = (self.vits_model.batched_decode(
                #         pred_semantic, pred_semantic_len, batch_phones, batch_phones_len,refer_audio_spec
                #     ))

                if speed_factor == 1.0:
                    # ## vits并行推理 method 2
                    pred_semantic_list = [item[-idx:] for item, idx in zip(pred_semantic_list, idx_list)]
                    upsample_rate = math.prod(self.vits_model.upsample_rates)
                    audio_frag_idx = [pred_semantic_list[i].shape[0]*2*upsample_rate for i in range(0, len(pred_semantic_list))]
                    audio_frag_end_idx = [ sum(audio_frag_idx[:i+1]) for i in range(0, len(audio_frag_idx))]
                    all_pred_semantic = torch.cat(pred_semantic_list).unsqueeze(0).unsqueeze(0).to(self.configs.device)
                    _batch_phones = torch.cat(batch_phones).unsqueeze(0).to(self.configs.device)
                    _batch_audio_fragment = (self.vits_model.decode(
                            all_pred_semantic, _batch_phones, refer_audio_spec, speed=speed_factor
                        ).detach()[0, 0, :])
                    audio_frag_end_idx.insert(0, 0)
                    batch_audio_fragment= [_batch_audio_fragment[audio_frag_end_idx[i-1]:audio_frag_end_idx[i]] for i in range(1, len(audio_frag_end_idx))]
                else:
                # ## vits串行推理
                    for i, idx in enumerate(idx_list):
                        phones = batch_phones[i].unsqueeze(0).to(self.configs.device)
                        _pred_semantic = (pred_semantic_list[i][-idx:].unsqueeze(0).unsqueeze(0))   # .unsqueeze(0)#mq要多unsqueeze一次
                        audio_fragment =(self.vits_model.decode(
                                _pred_semantic, phones, refer_audio_spec, speed=speed_factor
                            ).detach()[0, 0, :])
                        batch_audio_fragment.append(
                            audio_fragment
                        )  ###试试重建不带上prompt部分

                t5 = ttime()
                t_45 += t5 - t4
                if return_fragment:
                    print("%.3f\t%.3f\t%.3f\t%.3f" % (t1 - t0, t2 - t1, t4 - t3, t5 - t4))
                    yield self.audio_postprocess([batch_audio_fragment], 
                                                    self.configs.sampling_rate, 
                                                    None, 
                                                    speed_factor, 
                                                    False,
                                                    fragment_interval
                                                    )
                else:
                    audio.append(batch_audio_fragment)

                if self.stop_flag:
                    yield self.configs.sampling_rate, np.zeros(int(self.configs.sampling_rate),
                                                            dtype=np.int16)
                    return

            if not return_fragment:
                print("%.3f\t%.3f\t%.3f\t%.3f" % (t1 - t0, t2 - t1, t_34, t_45))
                if len(audio) == 0:
                    yield self.configs.sampling_rate, np.zeros(int(self.configs.sampling_rate),
                                                                dtype=np.int16)
                    return
                yield self.audio_postprocess(audio, 
                                                self.configs.sampling_rate, 
                                                batch_index_list, 
                                                speed_factor, 
                                                split_bucket,
                                                fragment_interval
                                                )

        except Exception as e:
            traceback.print_exc()
            # 必须返回一个空音频, 否则会导致显存不释放。
            yield self.configs.sampling_rate, np.zeros(int(self.configs.sampling_rate),
                                                            dtype=np.int16)
            # 重置模型, 否则会导致显存释放不完全。
            del self.t2s_model
            del self.vits_model
            self.t2s_model = None
            self.vits_model = None
            self.init_t2s_weights(self.configs.t2s_weights_path)
            self.init_vits_weights(self.configs.vits_weights_path)
            raise e
        finally:
            self.empty_cache()
    
    def empty_cache(self):
        try:
            gc.collect() # 触发gc的垃圾回收。避免内存一直增长。
            if "cuda" in str(self.configs.device):
                torch.cuda.empty_cache()
            elif str(self.configs.device) == "mps":
                torch.mps.empty_cache()
        except:
            pass 
        
    def audio_postprocess(self, 
                          audio:List[torch.Tensor], 
                          sr:int, 
                          batch_index_list:list=None, 
                          speed_factor:float=1.0, 
                          split_bucket:bool=True,
                          fragment_interval:float=0.3
                          )->Tuple[int, np.ndarray]:
        zero_wav = torch.zeros(
                        int(self.configs.sampling_rate * fragment_interval),
                        dtype=self.precision,
                        device=self.configs.device
                    )
        
        for i, batch in enumerate(audio):
            for j, audio_fragment in enumerate(batch):
                max_audio=torch.abs(audio_fragment).max()#简单防止16bit爆音
                if max_audio>1: audio_fragment/=max_audio
                audio_fragment:torch.Tensor = torch.cat([audio_fragment, zero_wav], dim=0)
                audio[i][j] = audio_fragment.cpu().numpy()
            
        
        if split_bucket:
            audio = self.recovery_order(audio, batch_index_list)
        else:
            # audio = [item for batch in audio for item in batch]
            audio = sum(audio, [])
            
            
        audio = np.concatenate(audio, 0)
        audio = (audio * 32768).astype(np.int16) 
        
        # try:
        #     if speed_factor != 1.0:
        #         audio = speed_change(audio, speed=speed_factor, sr=int(sr))
        # except Exception as e:
        #     print(f"Failed to change speed of audio: \n{e}")
        
        return sr, audio
            
        
        
       
def speed_change(input_audio:np.ndarray, speed:float, sr:int):
    # 将 NumPy 数组转换为原始 PCM 流
    raw_audio = input_audio.astype(np.int16).tobytes()

    # 设置 ffmpeg 输入流
    input_stream = ffmpeg.input('pipe:', format='s16le', acodec='pcm_s16le', ar=str(sr), ac=1)

    # 变速处理
    output_stream = input_stream.filter('atempo', speed)

    # 输出流到管道
    out, _ = (
        output_stream.output('pipe:', format='s16le', acodec='pcm_s16le')
        .run(input=raw_audio, capture_stdout=True, capture_stderr=True)
    )

    # 将管道输出解码为 NumPy 数组
    processed_audio = np.frombuffer(out, np.int16)

    return processed_audio


# This code is modified from https://github.com/mozillazg/pypinyin-g2pW

import pickle
import os

from pypinyin.constants import RE_HANS
from pypinyin.core import Pinyin, Style
from pypinyin.seg.simpleseg import simple_seg
from pypinyin.converter import UltimateConverter
from pypinyin.contrib.tone_convert import to_tone
from .onnx_api import G2PWOnnxConverter

current_file_path = os.path.dirname(__file__)
CACHE_PATH = os.path.join(current_file_path, "polyphonic.pickle")
PP_DICT_PATH = os.path.join(current_file_path, "polyphonic.rep")
PP_FIX_DICT_PATH = os.path.join(current_file_path, "polyphonic-fix.rep")


class G2PWPinyin(Pinyin):
    def __init__(self, model_dir='G2PWModel/', model_source=None,
                 enable_non_tradional_chinese=True,
                 v_to_u=False, neutral_tone_with_five=False, tone_sandhi=False, **kwargs):
        self._g2pw = G2PWOnnxConverter(
            model_dir=model_dir,
            style='pinyin',
            model_source=model_source,
            enable_non_tradional_chinese=enable_non_tradional_chinese,
        )
        self._converter = Converter(
            self._g2pw, v_to_u=v_to_u,
            neutral_tone_with_five=neutral_tone_with_five,
            tone_sandhi=tone_sandhi,
        )

    def get_seg(self, **kwargs):
        return simple_seg


class Converter(UltimateConverter):
    def __init__(self, g2pw_instance, v_to_u=False,
                 neutral_tone_with_five=False,
                 tone_sandhi=False, **kwargs):
        super(Converter, self).__init__(
            v_to_u=v_to_u,
            neutral_tone_with_five=neutral_tone_with_five,
            tone_sandhi=tone_sandhi, **kwargs)

        self._g2pw = g2pw_instance

    def convert(self, words, style, heteronym, errors, strict, **kwargs):
        pys = []
        if RE_HANS.match(words):
            pys = self._to_pinyin(words, style=style, heteronym=heteronym,
                                  errors=errors, strict=strict)
            post_data = self.post_pinyin(words, heteronym, pys)
            if post_data is not None:
                pys = post_data

            pys = self.convert_styles(
                pys, words, style, heteronym, errors, strict)

        else:
            py = self.handle_nopinyin(words, style=style, errors=errors,
                                      heteronym=heteronym, strict=strict)
            if py:
                pys.extend(py)

        return _remove_dup_and_empty(pys)

    def _to_pinyin(self, han, style, heteronym, errors, strict, **kwargs):
        pinyins = []

        g2pw_pinyin = self._g2pw(han)

        if not g2pw_pinyin:  # g2pw 不支持的汉字改为使用 pypinyin 原有逻辑
            return super(Converter, self).convert(
                         han, Style.TONE, heteronym, errors, strict, **kwargs)

        for i, item in enumerate(g2pw_pinyin[0]):
            if item is None:  # g2pw 不支持的汉字改为使用 pypinyin 原有逻辑
                py = super(Converter, self).convert(
                           han[i], Style.TONE, heteronym, errors, strict, **kwargs)
                pinyins.extend(py)
            else:
                pinyins.append([to_tone(item)])

        return pinyins


def _remove_dup_items(lst, remove_empty=False):
    new_lst = []
    for item in lst:
        if remove_empty and not item:
            continue
        if item not in new_lst:
            new_lst.append(item)
    return new_lst


def _remove_dup_and_empty(lst_list):
    new_lst_list = []
    for lst in lst_list:
        lst = _remove_dup_items(lst, remove_empty=True)
        if lst:
            new_lst_list.append(lst)
        else:
            new_lst_list.append([''])

    return new_lst_list


def cache_dict(polyphonic_dict, file_path):
    with open(file_path, "wb") as pickle_file:
        pickle.dump(polyphonic_dict, pickle_file)


def get_dict():
    if os.path.exists(CACHE_PATH):
        with open(CACHE_PATH, "rb") as pickle_file:
            polyphonic_dict = pickle.load(pickle_file)
    else:
        polyphonic_dict = read_dict()
        cache_dict(polyphonic_dict, CACHE_PATH)

    return polyphonic_dict


def read_dict():
    polyphonic_dict = {}
    with open(PP_DICT_PATH,encoding="utf-8") as f:
        line = f.readline()
        while line:
            key, value_str = line.split(':')
            value = eval(value_str.strip())
            polyphonic_dict[key.strip()] = value
            line = f.readline()
    with open(PP_FIX_DICT_PATH,encoding="utf-8") as f:
        line = f.readline()
        while line:
            key, value_str = line.split(':')
            value = eval(value_str.strip())
            polyphonic_dict[key.strip()] = value
            line = f.readline()
    return polyphonic_dict


def correct_pronunciation(word,word_pinyins):
    if word in pp_dict:
        word_pinyins = pp_dict[word]

    return word_pinyins


pp_dict = get_dict()



import asyncio
from collections import OrderedDict
from typing import TypeVar, Optional, Union, Tuple, List
import numpy as np
from starlette.websockets import WebSocket
import base64
import soundfile as sf
import aiohttp
import io
from gallama.data_classes.realtime_client_proto import (
    ConversationItem,
    ConversationItemDelete,
    ConversationItemTruncate,
)
from gallama.data_classes.realtime_server_proto import ConversationItemDeleted, ConversationItemTruncated, \
    MessageContentServer, ConversationItemMessageServer, ConversationItemServer, parse_conversation_item, \
    ConversationItemCreated
from gallama.data_classes.internal_ws import (
    WSInterSTT
)
from gallama.realtime.websocket_client import WebSocketClient
from difflib import SequenceMatcher
import uuid
from ..dependencies_server import get_server_logger

logger = get_server_logger()


T = TypeVar("T", bound=ConversationItemServer)


class MessageQueues:
    """Manages different message queues for the websocket system"""
    def __init__(self):
        # current audio will store into this array.
        # once audio commited, it will move to part of conversation_itm_od
        self.audio_buffer = np.array([], dtype=np.float32)

        # transcription of audio until before commit will be in self.transcription_buffer
        self.transcript_buffer = ""
        self.transcription_complete = False
        self.audio_commited = False
        self.lock_transcript_complete = asyncio.Lock()

        self.sample_rate = 24000    # assuming fixed at this rate
        self.verbose = False  # Control debug logging

        # all non audio events go here
        self.unprocessed = asyncio.Queue()

        # ordered dict of conversation item
        self.conversation_item_od: OrderedDict[str, T] = OrderedDict()

        self.response_queue = asyncio.Queue()
        self.audio_to_client = asyncio.Queue()
        self.response_counter = ""
        self.event_counter = ""
        self.item_counter = ""

        self.lock_conversation_item = asyncio.Lock()
        self.lock_response_counter = asyncio.Lock()
        self.lock_event_counter = asyncio.Lock()
        self.lock_item_counter = asyncio.Lock()

        self.lock_audio_buffer = asyncio.Lock()     # ensure that audio is sync with ws_stt
        self.lock_transcript_buffer = asyncio.Lock()     # ensure that audio is sync with ws_stt
        self.lock_audio_commited = asyncio.Lock()     # ensure that audio is sync with ws_stt

        self.vad_item_id = None
        # this is id to assign to vad_item_id while it is not commited yet
        # the reason to have this variable is because response will clear audio once it done
        self.vad_item_id_next_in_queue = None
        self.speech_start = None
        self.speech_end = None

        # keep track of when speech start and end for video
        self.speech_start_time: float = 0.0
        self.speech_start_time_in_queue: float = 0.0
        self.speech_end_time: float = 0.0
        # self.latest_item: Optional[ConversationItem] = None

        # self.uncommitted_audio_data: Optional[bytes] = None
        # self.uncommitted_text: Optional[str] = None

    async def next_event(self) -> str:
        """ return the next counter for event"""
        async with self.lock_event_counter:
            self.event_counter = str(uuid.uuid4().hex)
            return f"event_{self.event_counter}"

    async def next_resp(self) -> str:
        """ return the next counter for response"""
        async with self.lock_response_counter:
            self.response_counter = str(uuid.uuid4().hex)
            return f"resp_{self.response_counter}"

    async def next_item(self, return_current=False) -> Union[str,Tuple[str,Union[str, None]]]:
        """ return the next counter for response"""
        async with self.lock_item_counter:
            self.item_counter = str(uuid.uuid4().hex)
            if return_current:
                if not self.conversation_item_od:
                    return f"item_{self.item_counter}", None
                else:
                    return f"item_{self.item_counter}", next(reversed(self.conversation_item_od.keys()))
            else:
                return f"item_{self.item_counter}"

    async def current_item_id(self) -> str | None:
        """ return the current item id"""
        if not self.conversation_item_od:
            return None
        else:
            return next(reversed(self.conversation_item_od.keys()))     # id of the latest message


    async def get_previous_item_id(self, message_id: str) -> Optional[str]:
        """Get the ID of the message that comes before the given message_id. This only apply to committed messages"""
        async with self.lock_conversation_item:
            previous_id = None
            for id in self.conversation_item_od.keys():
                if id == message_id:
                    return previous_id
                previous_id = id
            return None  # Message ID not found

    async def append_unprocessed_audio(self, base64audio: str, ws_stt: WebSocketClient, audio_float: Optional[np.ndarray] = None):
        """
        Appends base64 encoded audio data to the audio buffer after decoding.
        float32 normalized audio data in numpy array
        Args:
            base64audio (str): Base64 encoded audio string
            ws_stt (WebSocketClient): the websocket client for ws_stt

        Returns:
            None
        """
        try:
            if not audio_float.any():
                # Decode base64 string to bytes
                audio_bytes = base64.b64decode(base64audio)

                # Convert bytes to numpy array of int16 values
                audio_data = np.frombuffer(audio_bytes, dtype=np.int16)

                # Convert int16 to float32 (normalize to [-1.0, 1.0] range)
                audio_float = audio_data.astype(np.float32) / 32768.0

            # Append to existing buffer
            async with self.lock_audio_buffer:
                self.audio_buffer = np.concatenate([self.audio_buffer, audio_float])

            # send the audio to ws_stt
            await ws_stt.send_pydantic_message(
                WSInterSTT(
                    type="stt.add_sound_chunk",
                    sound=base64audio,
                )
            )
        except Exception as e:
            # Log error and optionally raise it depending on your error handling needs
            logger.error(f"Error processing audio data: {str(e)}")
            raise


    async def commit_unprocessed_audio(self, ws_stt: WebSocketClient, item_id: str = None, skip_sound_done:bool = False):
        """
        Signal to ws_stt that the audio buffer has been committed.

        If a specific item id to be used, it can be passed in

        Returns:
            None
        """
        try:
            if not skip_sound_done:
                # send the ws_stt the commit signal
                async with self.lock_audio_buffer:
                    await ws_stt.send_pydantic_message(WSInterSTT(type="stt.sound_done"))

            async with self.lock_audio_commited:
                self.audio_commited = True

            self.vad_item_id = item_id
        except Exception as e:
            # Log error and optionally raise it depending on your error handling needs
            logger.error(f"Error processing audio data: {str(e)}")
            raise

    async def clear_unprocessed_audio(self, ws_stt: WebSocketClient):
        """
        Signal to ws_stt that the audio buffer has been committed.
        After stt clear the buffer on its side, it will send the acknowledgement.
        There will be another job that based on the acknowledgement to clear the transcription buffer
        Returns:
            None
        """
        try:
            # send the ws_stt the clear signal
            async with self.lock_audio_buffer:
                await ws_stt.send_pydantic_message(WSInterSTT(type="stt.clear_buffer"))
                # reset audio
                await self.reset_audio()
        except Exception as e:
            # Log error and optionally raise it depending on your error handling needs
            logger.error(f"Error processing audio data: {str(e)}")
            raise

    async def append_transcription(self, transcription_chunk):
        async with self.lock_transcript_buffer:
            self.transcript_buffer += transcription_chunk


    async def mark_transcription_done(self):
        logger.info(f"in MessageQueue: transcription done")
        async with self.lock_transcript_complete:
            self.transcription_complete = True

    async def clear_transcription(self):
        async with self.lock_transcript_buffer:
            self.transcript_buffer = ""
        async with self.lock_transcript_complete:
            self.transcription_complete = False

    async def clear_audio_buffer(self):
        async with self.lock_audio_buffer:
            self.audio_buffer = np.array([], dtype=np.float32)

    def calculate_audio_duration(self):
        """Calculate the duration of the audio in the audio_buffer in seconds."""
        if self.audio_buffer.any():
            duration = len(self.audio_buffer) / self.sample_rate
            return duration
        else:
            return 0

    async def reset_audio(self):
        try:
            clear_audio_task = self.clear_audio_buffer()
            clear_transcription_buffer = self.clear_transcription()
            self.vad_item_id = None

            await asyncio.gather(clear_audio_task, clear_transcription_buffer)

            async with self.lock_audio_commited:
                self.audio_commited = False

            return True
        except Exception as e:
            logger.error(f"Error reset_audio data: {str(e)}")
            return False

    async def audio_exist(self):
        """ return True or False if there is any audio at all in the conversation"""
        async with self.lock_audio_buffer:
            if self.audio_buffer.any():
                return True
            else:
                return False



    async def reset_after_response(self):
        # for now only audio need to reset
        _ = await self.reset_audio()
        logger.info(f"reset after response completed")


    async def wait_for_transcription_done(self, use_vad: bool = False):
        """wait for audio to commit with 10 second timeout

        Returns:
            bool: True if transcription completed, False if timeout occurred
        """
        try:
            # just exit if use_vad but no speech detected
            if use_vad and self.vad_item_id is None:
                return False, None

            # if not use vad, but audio not commited
            if not use_vad and not self.audio_commited:
                return False, None

            async def _wait():
                while not self.transcription_complete:
                    await asyncio.sleep(0.05)
                return True

            await asyncio.wait_for(_wait(), timeout=10.0)
            logger.info(f"transcription done")
            return True, self.vad_item_id
        except asyncio.TimeoutError:
            return False, None
        except Exception as e:
            raise Exception(f"Error in wait_for_transcription_done: {str(e)}")

    async def update_conversation_item_ordered_dict_client(
        self,
        item: ConversationItemServer        # item to create
    ):
        """
        Add an item to the conversation item list.
        If it is a new item, return a ConversationItemCreated object for onward return to front end
        """
        async with self.lock_conversation_item:
            # update to history
            if isinstance(item, ConversationItem):
                self.conversation_item_od[item.id] = item


    async def update_conversation_item_ordered_dict(
        self,
        ws_client: WebSocket,               # web socket for client
        ws_llm: WebSocketClient,            # web socket for llm
        item: ConversationItemServer,        # item to create
    ):
        """
        Add an item to the conversation item list.
        If it is a new item, return a ConversationItemCreated object for onward return to front end
        """
        try:
            new_event_id = await self.next_event()
            previous_item_id = await self.get_previous_item_id(item.id)
            # if no previous item, meaning this is new item
            if not previous_item_id:
                previous_item_id = await self.current_item_id()

            item_to_send = item

            # send client update about new item created
            async with self.lock_conversation_item:
                if not item.id:
                    item.id = await self.next_item()

                # send to ws_llm to sync with this item
                stripped_item = item.strip_audio(mode="remove")
                await ws_llm.send_pydantic_message(stripped_item)

                if item.id not in self.conversation_item_od.keys():
                    item_to_send = ConversationItemCreated(**{
                        "event_id": new_event_id,
                        "type": "conversation.item.created",
                        "previous_item_id": previous_item_id,
                        "item": stripped_item.model_dump(exclude_none=True)
                    }).model_dump(exclude_none=True)

                # update to history with server type item
                item_server = parse_conversation_item(item.model_dump())
                self.conversation_item_od[item.id] = item_server

                # convert audio to base64 & send to client
                if not isinstance(item_to_send, dict):
                    item_to_send = item_to_send.model_dump(exclude_none=True)

                await ws_client.send_json(item_to_send)
        except Exception as e:
            logger.error(f"Error in update_conversation_item_ordered_dict: {str(e)}")
            raise

    async def update_conversation_item_with_assistant_response(
        self,
        item_id: str,
        _type: str,
        text: str,
        transcript: str,
        audio: np.ndarray,
        function_calling_name: str,
        function_calling_arguments: str,
        function_calling_tool_id: str,
        ws_llm: WebSocketClient,  # web socket for llm
    ):
        """
        Add an item to the conversation item list.
        If it is a new item, return a ConversationItemCreated object for onward return to front end
        """
        try:
            # update value in place to prevent moving its position

            if not function_calling_name:   # non function calling
                self.conversation_item_od[item_id].content = [
                    MessageContentServer(
                        type=_type,
                        text=text,
                        transcript=transcript,
                        audio=audio,
                    )
                ]
                stripped_item = self.conversation_item_od[item_id].strip_audio()
            else:
                self.conversation_item_od[item_id].call_id = function_calling_tool_id
                self.conversation_item_od[item_id].name = function_calling_name
                self.conversation_item_od[item_id].arguments = function_calling_arguments
                stripped_item = self.conversation_item_od[item_id]

            # send to ws_llm to sync with this item
            await ws_llm.send_pydantic_message(stripped_item)

        except Exception as e:
            logger.error(f"Error in update_conversation_item_with_assistant_response: {str(e)}")
            raise


    async def delete_conversation_item(
            self,
            ws_client: WebSocket,
            ws_llm: WebSocketClient,
            item_id: str,
    ):
        """
        Delete an item from the conversation item list.
        Update previous_item_id for the next item.
        Send this item to ws_llm as well to sync the state.
        Notify the client about the deletion via a ConversationItemDeleted event.

        Args:
            ws_client (WebSocket): The websocket connection to the client
            ws_llm (WebSocketClient): The websocket connection to the LLM service
            item_id (str): The ID of the item to delete

        Raises:
            Exception: If there's an error during the deletion process
        """
        try:
            new_event_id = await self.next_event()

            async with self.lock_conversation_item:
                # Check if item exists before trying to delete
                if item_id not in self.conversation_item_od:
                    logger.warning(f"Attempted to delete non-existent item: {item_id}")
                    return

                # Remove the item from the ordered dictionary
                del self.conversation_item_od[item_id]

                # Create deletion event for the client
                delete_event = ConversationItemDeleted(
                    event_id=new_event_id,
                    type="conversation.item.deleted",
                    item_id=item_id
                )

                # Send deletion event to the client
                await ws_client.send_json(delete_event.model_dump())

                # Send deletion event to the LLM to sync state
                await ws_llm.send_pydantic_message(delete_event)

        except Exception as e:
            logger.error(f"Error in delete_conversation_item: {str(e)}")
            raise


    def get_last_conversation_item(self):
        return next(reversed(self.conversation_item_od.items()))

    @staticmethod
    def find_truncated_match(partial_text, full_transcript):
        # Store original words for output
        full_words_original = full_transcript.split()

        # Convert to lowercase only for matching
        partial_text_lower = partial_text.lower().strip()
        full_transcript_lower = full_transcript.lower().strip()

        partial_words = partial_text_lower.split()
        partial_length = len(partial_words)

        full_words_lower = full_transcript_lower.split()

        best_ratio = 0
        best_end_pos = partial_length

        for end_pos in range(partial_length - 2, partial_length + 3):
            if end_pos <= 0 or end_pos >= len(full_words_lower):
                continue

            current_chunk = ' '.join(full_words_lower[:end_pos])
            sequence_ratio = SequenceMatcher(None, partial_text_lower, current_chunk).ratio()

            if sequence_ratio > best_ratio:
                best_ratio = sequence_ratio
                best_end_pos = end_pos

        return {
            'matched_text': ' '.join(full_words_original[:best_end_pos]),  # Use original text
            'next_words': ' '.join(full_words_original[best_end_pos:best_end_pos + 3]),  # Use original text
            'confidence': best_ratio
        }

    async def _debug_audio_diagnostics(self, audio_data: np.ndarray, truncated_audio: np.ndarray, end_sample: int):
        """Log detailed audio diagnostics and save debug files when verbose is True."""
        if not self.verbose:
            return

        logger.info("=== AUDIO DIAGNOSTIC INFO ===")
        logger.info(f"Original audio data shape: {audio_data.shape}, dtype: {audio_data.dtype}")
        logger.info(f"Sample rate: {self.sample_rate}")
        logger.info(f"Input duration: {len(audio_data) / self.sample_rate:.2f} seconds")
        logger.info(f"Expected samples for 1 second: {self.sample_rate}")
        logger.info(f"Audio min value: {np.min(audio_data)}, max value: {np.max(audio_data)}")
        logger.info(f"Truncating at sample {end_sample}")
        logger.info(f"Calculated duration after truncation: {end_sample / self.sample_rate:.2f} seconds")
        logger.info(f"Truncated audio shape: {truncated_audio.shape}, dtype: {truncated_audio.dtype}")
        logger.info(f"Truncated audio min value: {np.min(truncated_audio)}, max value: {np.max(truncated_audio)}")

        if self.verbose:
            await self._save_debug_audio_files(truncated_audio)

    async def _save_debug_audio_files(self, truncated_audio: np.ndarray):
        """Save audio files for debugging when verbose is True."""
        try:
            # Save as 16-bit PCM WAV at 16kHz
            debug_path = "/home/remichu/work/ML/gallama/experiment/debug_audio.wav"
            sf.write(debug_path, truncated_audio, self.sample_rate, subtype='PCM_16')
            logger.info(f"Successfully saved debug audio file to: {debug_path}")

            # Save another copy at 44.1kHz for comparison
            debug_path_441 = "/home/remichu/work/ML/gallama/experiment/debug_audio_44100.wav"
            import samplerate

            # Resample to 44.1kHz
            audio_441 = samplerate.resample(truncated_audio, 44100 / self.sample_rate, 'sinc_best')
            sf.write(debug_path_441, audio_441, 44100, subtype='PCM_16')
            logger.info(f"Saved 44.1kHz version to: {debug_path_441}")

        except Exception as e:
            logger.error(f"Error saving debug audio files: {str(e)}")

    async def _prepare_truncated_audio_buffer(self, audio_data: np.ndarray, audio_end_ms: int) -> tuple[np.ndarray, io.BytesIO]:
        """Prepare truncated audio data and buffer for transcription."""
        # Convert milliseconds to samples
        end_sample = int((audio_end_ms / 1000) * self.sample_rate)

        # Truncate the audio data
        truncated_audio = audio_data[:end_sample]

        # Ensure audio data is in float32 format
        if truncated_audio.dtype != np.float32:
            if truncated_audio.dtype == np.int16:
                truncated_audio = truncated_audio.astype(np.float32) / 32768.0
            elif truncated_audio.dtype == np.int32:
                truncated_audio = truncated_audio.astype(np.float32) / 2147483648.0
            else:
                truncated_audio = truncated_audio.astype(np.float32)

        # Log diagnostics if verbose
        if self.verbose:
            await self._debug_audio_diagnostics(audio_data, truncated_audio, end_sample)

        # Prepare buffer for API call
        buffer = io.BytesIO()
        sf.write(buffer, truncated_audio, self.sample_rate, format='wav', subtype='PCM_16')
        buffer.seek(0)

        return truncated_audio, buffer

    async def _get_transcription(self, audio_buffer: io.BytesIO) -> str:
        """Get transcription for the truncated audio."""
        form_data = aiohttp.FormData()
        form_data.add_field('file', audio_buffer, filename='audio.wav', content_type='audio/wav')
        form_data.add_field('model', 'whisper-1')
        form_data.add_field('response_format', 'json')
        form_data.add_field('temperature', str(0.0))
        form_data.add_field('timestamp_granularities', 'segment')

        async with aiohttp.ClientSession() as session:
            async with session.post(
                    'http://localhost:8001/v1/audio/transcriptions',
                    data=form_data
            ) as response:
                if response.status == 200:
                    response_data = await response.json()
                    transcription = response_data.get('text')
                    if transcription is None:
                        raise Exception("Transcription response missing 'text' field")
                    return transcription
                else:
                    error_detail = await response.text()
                    raise Exception(f"Transcription failed with status {response.status}: {error_detail}")

    async def truncate_conversation_item(
        self,
        ws_client: WebSocket,
        ws_llm: WebSocketClient,
        event: ConversationItemTruncate,
        user_interrupt_token: str = "",
        offset_timing: int = 0
    ):
        """Truncate an audio item from the conversation item list."""
        try:
            async with self.lock_conversation_item:
                # item truncation should hold a lock until it complete

                new_event_id = await self.next_event()
                item_id = event.item_id
                audio_end_ms = event.audio_end_ms

                # Send acknowledgement to client
                async def send_client_acknowledgement():
                    return await ws_client.send_json(ConversationItemTruncated(
                        event_id=new_event_id,
                        type="conversation.item.truncated",
                        item_id=item_id,
                        content_index=0,
                        audio_end_ms=audio_end_ms
                    ).model_dump())

                async def _update_internal_state():
                    # Get original audio data
                    audio_data = self.conversation_item_od[item_id].content[0].audio

                    # Prepare truncated audio and buffer
                    # offset another 25ms for latency
                    audio_end_ms_offset = max(audio_end_ms - offset_timing - 25, 0)

                    if audio_end_ms_offset == 0:
                        logger.info("No need to truncate audio")
                        return
                    else:
                        logger.info(f"Truncate audio from {audio_end_ms} to {audio_end_ms_offset}")

                    truncated_audio, audio_buffer = await self._prepare_truncated_audio_buffer(audio_data, audio_end_ms_offset)

                    # Get new transcription
                    new_transcription = await self._get_transcription(audio_buffer)

                    # Fuzzy match the transcription
                    new_transcription_matched = self.find_truncated_match(
                        partial_text=new_transcription,
                        full_transcript=self.conversation_item_od[item_id].content[0].transcript,
                    )
                    new_transcription = new_transcription_matched["matched_text"]

                    logger.info(f"Original transcription: {self.conversation_item_od[item_id].content[0].transcript}")
                    logger.info(f"After truncated transcription: {new_transcription}")

                    # Update conversation item
                    current_item = self.conversation_item_od[item_id].content[0]
                    current_item.audio = truncated_audio
                    current_item.text = new_transcription + user_interrupt_token
                    current_item.transcript = new_transcription + user_interrupt_token

                    # Sync with LLM
                    stripped_item = self.conversation_item_od[item_id].strip_audio()
                    await ws_llm.send_pydantic_message(stripped_item)

                # Execute both tasks concurrently
                await asyncio.gather(send_client_acknowledgement(), _update_internal_state())

        except Exception as e:
            logger.error(f"Error in truncate_conversation_item: {str(e)}")
            raise

    def reset(self):
        self.audio_buffer = np.array([], dtype=np.float32)

        # transcription of audio until before commit will be in self.transcription_buffer
        self.transcript_buffer = ""
        self.transcription_complete = False
        self.audio_commited = False
        self.lock_transcript_complete = asyncio.Lock()

        self.sample_rate = 24000  # assuming fixed at this rate
        self.verbose = False  # Control debug logging

        # all non audio events go here
        self.unprocessed = asyncio.Queue()

        # ordered dict of conversation item
        self.conversation_item_od: OrderedDict[str, T] = OrderedDict()

        self.response_queue = asyncio.Queue()
        self.audio_to_client = asyncio.Queue()
        self.response_counter = ""
        self.event_counter = ""
        self.item_counter = ""

        self.lock_conversation_item = asyncio.Lock()
        self.lock_response_counter = asyncio.Lock()
        self.lock_event_counter = asyncio.Lock()
        self.lock_item_counter = asyncio.Lock()

        self.lock_audio_buffer = asyncio.Lock()  # ensure that audio is sync with ws_stt
        self.lock_transcript_buffer = asyncio.Lock()  # ensure that audio is sync with ws_stt
        self.lock_audio_commited = asyncio.Lock()  # ensure that audio is sync with ws_stt

        self.vad_item_id = None
        self.speech_start = None
        self.speech_end = None


import numpy as np

import gallama.realtime.message_queue


# This function is obtained from librosa.
def get_rms(
    y,
    frame_length=2048,
    hop_length=512,
    pad_mode="constant",
):
    padding = (int(frame_length // 2), int(frame_length // 2))
    y = np.pad(y, padding, mode=pad_mode)

    axis = -1
    # put our new within-frame axis at the end for now
    out_strides = y.strides + tuple([y.strides[axis]])
    # Reduce the shape on the framing axis
    x_shape_trimmed = list(y.shape)
    x_shape_trimmed[axis] -= frame_length - 1
    out_shape = tuple(x_shape_trimmed) + tuple([frame_length])
    xw = np.lib.stride_tricks.as_strided(y, shape=out_shape, strides=out_strides)
    if axis < 0:
        target_axis = axis - 1
    else:
        target_axis = axis + 1
    xw = np.moveaxis(xw, -1, target_axis)
    # Downsample along the target axis
    slices = [slice(None)] * xw.ndim
    slices[axis] = slice(0, None, hop_length)
    x = xw[tuple(slices)]

    # Calculate power
    power = np.mean(np.abs(x) ** 2, axis=-2, keepdims=True)

    return np.sqrt(power)


class Slicer:
    def __init__(
        self,
        sr: int,
        threshold: float = -40.0,
        min_length: int = 5000,
        min_interval: int = 300,
        hop_size: int = 20,
        max_sil_kept: int = 5000,
    ):
        if not min_length >= min_interval >= hop_size:
            raise ValueError(
                "The following condition must be satisfied: min_length >= min_interval >= hop_size"
            )
        if not max_sil_kept >= hop_size:
            raise ValueError(
                "The following condition must be satisfied: max_sil_kept >= hop_size"
            )
        min_interval = sr * min_interval / 1000
        self.threshold = 10 ** (threshold / 20.0)
        self.hop_size = round(sr * hop_size / 1000)
        self.win_size = min(round(min_interval), 4 * self.hop_size)
        self.min_length = round(sr * min_length / 1000 / self.hop_size)
        self.min_interval = round(min_interval / self.hop_size)
        self.max_sil_kept = round(sr * max_sil_kept / 1000 / self.hop_size)

    def _apply_slice(self, waveform, begin, end):
        if len(waveform.shape) > 1:
            return waveform[
                :, begin * self.hop_size : min(waveform.shape[1], end * self.hop_size)
            ]
        else:
            return waveform[
                begin * self.hop_size : min(waveform.shape[0], end * self.hop_size)
            ]

    # @timeit
    def slice(self, waveform):
        if len(waveform.shape) > 1:
            samples = waveform.mean(axis=0)
        else:
            samples = waveform
        if samples.shape[0] <= self.min_length:
            return [waveform]
        rms_list = get_rms(
            y=samples, frame_length=self.win_size, hop_length=self.hop_size
        ).squeeze(0)
        sil_tags = []
        silence_start = None
        clip_start = 0
        for i, rms in enumerate(rms_list):
            # Keep looping while frame is silent.
            if rms < self.threshold:
                # Record start of silent frames.
                if silence_start is None:
                    silence_start = i
                continue
            # Keep looping while frame is not silent and silence start has not been recorded.
            if silence_start is None:
                continue
            # Clear recorded silence start if interval is not enough or clip is too short
            is_leading_silence = silence_start == 0 and i > self.max_sil_kept
            need_slice_middle = (
                i - silence_start >= self.min_interval
                and i - clip_start >= self.min_length
            )
            if not is_leading_silence and not need_slice_middle:
                silence_start = None
                continue
            # Need slicing. Record the range of silent frames to be removed.
            if i - silence_start <= self.max_sil_kept:
                pos = rms_list[silence_start : i + 1].argmin() + silence_start
                if silence_start == 0:
                    sil_tags.append((0, pos))
                else:
                    sil_tags.append((pos, pos))
                clip_start = pos
            elif i - silence_start <= self.max_sil_kept * 2:
                pos = rms_list[
                    i - self.max_sil_kept : silence_start + self.max_sil_kept + 1
                ].argmin()
                pos += i - self.max_sil_kept
                pos_l = (
                    rms_list[
                        silence_start : silence_start + self.max_sil_kept + 1
                    ].argmin()
                    + silence_start
                )
                pos_r = (
                    rms_list[i - self.max_sil_kept : i + 1].argmin()
                    + i
                    - self.max_sil_kept
                )
                if silence_start == 0:
                    sil_tags.append((0, pos_r))
                    clip_start = pos_r
                else:
                    sil_tags.append((min(pos_l, pos), max(pos_r, pos)))
                    clip_start = max(pos_r, pos)
            else:
                pos_l = (
                    rms_list[
                        silence_start : silence_start + self.max_sil_kept + 1
                    ].argmin()
                    + silence_start
                )
                pos_r = (
                    rms_list[i - self.max_sil_kept : i + 1].argmin()
                    + i
                    - self.max_sil_kept
                )
                if silence_start == 0:
                    sil_tags.append((0, pos_r))
                else:
                    sil_tags.append((pos_l, pos_r))
                clip_start = pos_r
            silence_start = None
        # Deal with trailing silence.
        total_frames = rms_list.shape[0]
        if (
            silence_start is not None
            and total_frames - silence_start >= self.min_interval
        ):
            silence_end = min(total_frames, silence_start + self.max_sil_kept)
            pos = rms_list[silence_start : silence_end + 1].argmin() + silence_start
            sil_tags.append((pos, total_frames + 1))
        # Apply and return slices.
        ####音频+起始时间+终止时间
        if len(sil_tags) == 0:
            return [[waveform,0,int(total_frames*self.hop_size)]]
        else:
            chunks = []
            if sil_tags[0][0] > 0:
                chunks.append([self._apply_slice(waveform, 0, sil_tags[0][0]),0,int(sil_tags[0][0]*self.hop_size)])
            for i in range(len(sil_tags) - 1):
                chunks.append(
                    [self._apply_slice(waveform, sil_tags[i][1], sil_tags[i + 1][0]),int(sil_tags[i][1]*self.hop_size),int(sil_tags[i + 1][0]*self.hop_size)]
                )
            if sil_tags[-1][1] < total_frames:
                chunks.append(
                    [self._apply_slice(waveform, sil_tags[-1][1], total_frames),int(sil_tags[-1][1]*self.hop_size),int(total_frames*self.hop_size)]
                )
            return chunks


def main():
    import os.path
    from argparse import ArgumentParser

    import librosa
    import soundfile

    parser = ArgumentParser()
    parser.add_argument("audio", type=str, help="The audio to be sliced")
    parser.add_argument(
        "--out", type=str, help="Output directory of the sliced audio clips"
    )
    parser.add_argument(
        "--db_thresh",
        type=float,
        required=False,
        default=-40,
        help="The dB threshold for silence detection",
    )
    parser.add_argument(
        "--min_length",
        type=int,
        required=False,
        default=5000,
        help="The minimum milliseconds required for each sliced audio clip",
    )
    parser.add_argument(
        "--min_interval",
        type=int,
        required=False,
        default=300,
        help="The minimum milliseconds for a silence part to be sliced",
    )
    parser.add_argument(
        "--hop_size",
        type=int,
        required=False,
        default=10,
        help="Frame length in milliseconds",
    )
    parser.add_argument(
        "--max_sil_kept",
        type=int,
        required=False,
        default=500,
        help="The maximum silence length kept around the sliced clip, presented in milliseconds",
    )
    args = parser.parse_args()
    out = args.out
    if out is None:
        out = os.path.dirname(os.path.abspath(args.audio))
    audio, sr = librosa.load(args.audio, sr=None, mono=False)
    slicer = Slicer(
        sr=sr,
        threshold=args.db_thresh,
        min_length=args.min_length,
        min_interval=args.min_interval,
        hop_size=args.hop_size,
        max_sil_kept=args.max_sil_kept,
    )
    chunks = slicer.slice(audio)
    if not os.path.exists(out):
        os.makedirs(out)
    for i, chunk in enumerate(chunks):
        if len(chunk.shape) > 1:
            chunk = gallama.realtime.message_queue.T
        soundfile.write(
            os.path.join(
                out,
                f"%s_%d.wav"
                % (os.path.basename(args.audio).rsplit(".", maxsplit=1)[0], i),
            ),
            chunk,
            sr,
        )


if __name__ == "__main__":
    main()
